# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import csv

from io import StringIO

from MAPI import (
    TBL_ALL_COLUMNS, TABLE_SORT_ASCEND, TABLE_SORT_DESCEND, TBL_BATCH
)
from MAPI.Defs import PpropFindProp
from MAPI.Struct import MAPIErrorNotFound, SSort, SSortOrderSet

from .defs import REV_TAG

from . import property_ as _prop

class Table(object):
    """Table class

    Low-level abstraction for MAPI tables.
    """

    def __init__(self, server, mapiobj, mapitable, proptag=None,
                 restriction=None, order=None, columns=None):
        self.server = server
        self.mapiobj = mapiobj
        self.mapitable = mapitable
        self.proptag = proptag
        if columns:
            mapitable.SetColumns(columns, 0)
        else:
            # some columns are hidden by default TODO result (if at all)
            # depends on table implementation
            cols = mapitable.QueryColumns(TBL_ALL_COLUMNS)
            cols = cols or mapitable.QueryColumns(0) # fall-back
            mapitable.SetColumns(cols, 0)
        if restriction:
            self.mapitable.Restrict(restriction.mapiobj, TBL_BATCH)

    @property
    def header(self):
        """Return all table column names."""
        return [str(REV_TAG.get(c, hex(c))) \
            for c in self.mapitable.QueryColumns(0)]

    def rows(self, batch_size=100, page_start=None, page_limit=None):
        """Return all table rows."""
        offset = 0
        if page_limit is not None:
            batch_size = page_limit
        if page_start is not None:
            self.mapitable.SeekRow(0, page_start)
        try:
            while True:
                result = self.mapitable.QueryRows(batch_size, offset)
                if not result:
                    break

                for row in result:
                    yield [_prop.Property(self.mapiobj, c) for c in row]

                if page_limit is not None:
                    break
                offset += batch_size

        except MAPIErrorNotFound:
            pass

    @property
    def count(self):
        """Return table row count."""
        return self.mapitable.GetRowCount(0)

    def dict_rows(self, batch_size=100):
        """Return all table rows as dictionaries."""
        for row in self.rows(batch_size):
            yield dict((p.proptag, p.mapiobj.Value) for p in row)

    # TODO: apply batch_size as shown above
    def dict_(self, key, value):
        d = {}
        for row in self.mapitable.QueryRows(2147483647, 0):
            d[PpropFindProp(row, key).Value] = PpropFindProp(row, value).Value
        return d

    def index(self, key):
        """Return key->row dictionary keyed on given column (proptag)."""
        d = {}
        for row in self.mapitable.QueryRows(2147483647, 0):
            d[PpropFindProp(row, key).Value] = \
                dict((c.ulPropTag, c.Value) for c in row)
        return d

    def data(self, header=False):
        """Return list per row with textual representation for each element."""
        data = [[p.strval for p in row] for row in self.rows()]
        if header:
            data = [self.header] + data
        return data

    def text(self, borders=False):
        """Return textual table representation."""
        result = []
        data = self.data(header=True)
        colsizes = [max(len(d[i]) for d in data) for i in range(len(data[0]))]
        for d in data:
            line = []
            for size, c in zip(colsizes, d):
                line.append(c.ljust(size))
            result.append(' '.join(line))
        return '\n'.join(result)

    def csv(self, *args, **kwargs):
        """Return CSV data for table. All arguments are passed to
        csv.writer."""
        csvfile = StringIO()
        writer = csv.writer(csvfile, *args, **kwargs)
        writer.writerows(self.data(header=True))
        return csvfile.getvalue()

    def sort(self, tags):
        """Sort table.

        :param tags: Tag(s) on which to sort.
        """
        if not isinstance(tags, tuple):
            tags = (tags,)

        ascend_descend = [SSort(abs(tag),
            TABLE_SORT_DESCEND if tag < 0 else TABLE_SORT_ASCEND)
                for tag in tags]

        self.mapitable.SortTable(SSortOrderSet(ascend_descend, 0, 0), 0)

    def __iter__(self):
        return self.rows()

    def __unicode__(self):
        tablename = None
        if self.proptag:
            tablename = REV_TAG.get(self.proptag)
        return 'Table(%s)' % tablename

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

import codecs

from .address import Address
from .pidlid import (
    PidLidDistributionListMembers,
    PidLidDistributionListOneOffMembers
)

WRAPPED_ENTRYID_PREFIX = \
    codecs.decode('00000000C091ADD3519DCF11A4A900AA0047FAA4', 'hex')
WRAPPED_EID_TYPE_MASK = 0xf
WRAPPED_EID_TYPE_CONTACT = 3
WRAPPED_EID_TYPE_PERSONAL_DISTLIST = 4

class DistList(object):
    """DistList class

    A distribution list contains users, groups and/or sub distribution lists.
    """

    def __init__(self, item):
        self.item = item

    def members(self, expand=True):
        """Return all distribution list members,
        as :class:`addresses <Address>` and/or (sub)
        :class:`distribution lists <DistList>`.

        :param expand: expand sub distribution lists (optional)
        """
        members = self.item.get(PidLidDistributionListMembers)
        oneoffs = self.item.get(PidLidDistributionListOneOffMembers)

        for i, (member, oneoff) in enumerate(zip(members, oneoffs)):
            pos = len(WRAPPED_ENTRYID_PREFIX)
            prefix, flags, rest = member[:pos], member[pos], member[pos+1:]

            if (prefix == WRAPPED_ENTRYID_PREFIX and \
                (flags & WRAPPED_EID_TYPE_MASK) == \
                    WRAPPED_EID_TYPE_PERSONAL_DISTLIST):

                item = self.item.store.item(codecs.encode(rest, 'hex'))

                if expand:
                    for member2 in item.distlist.members(expand=True):
                        yield member2
                else:
                    yield DistList(item)

            else:
                yield Address(server=self.item.server, oneoff=oneoff)

    # TODO add, delete members

    def __unicode__(self):
        return 'DistList(%s)' % self.item.name

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-only
import pysolr

"""
index and query mapi fields via python-pysolr

note how simple this is compared to the xapian plugin.


"""

class Plugin:
    def __init__(self, index_path, log):
        self.log = log
        self.solr = pysolr.Solr(index_path)
        self.data = []

    def extract_terms(self, text):
        return text.split()

    def search(self, server_guid, store_guid, folder_ids, fields_terms, query, log):
        log.info('performing query: %s', query)
        return [r['docid'] for r in self.solr.search(query, fl=['docid'])]

    def suggest(self, server_guid, store_guid, terms, orig, log):
        return orig

    def update(self, doc):
        self.data.append(doc)

    def delete(self, doc):
        self.commit()
        self.solr.delete(q='sourcekey:%s' % doc['sourcekey'])

    def commit(self):
        if self.data:
            try:
                self.solr.add(self.data)
            finally:
                self.data = []
                self.charcount = 0

    def reindex(self, server_guid, store_guid):
        pass



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import MAPI_UNICODE
from MAPI.Struct import (
    ECGROUP, MAPIErrorNotFound, MAPIErrorInvalidParameter,
    MAPIErrorCollision
)
from .properties import Properties
from .errors import NotFoundError, DuplicateError
from .compat import benc as _benc

try:
    from . import server as _server
except ImportError: # pragma: no cover
    _server = sys.modules[__package__ + '.server']
try:
    from . import user as _user
except ImportError: # pragma: no cover
    _user = sys.modules[__package__ + '.user']

class Group(Properties):
    """Group class

    A group is a collection of :class:`users <User>` and/or
    subgroups of :class:`users <User>`.
    """

    def __init__(self, name, server=None):
        self.server = server or \
            _server.Server(_skip_check=True, parse_args=False)
        self._name = str(name)
        try:
            self._ecgroup = self.server.sa.GetGroup(
                self.server.sa.ResolveGroupName(self._name, MAPI_UNICODE),
                MAPI_UNICODE
            )
        except (MAPIErrorNotFound, MAPIErrorInvalidParameter):
            raise NotFoundError("no such group '%s'" % name)

        self._mapiobj = None

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if not self._mapiobj:
            self._mapiobj = self.server.mapisession.OpenEntry(
                self._ecgroup.GroupID, None, 0)
        return self._mapiobj

    @property
    def groupid(self):
        """Group id."""
        return _benc(self._ecgroup.GroupID)

    def users(self): # TODO recurse?
        """Return all :class:`users <User>` in group."""
        return self.members(groups=False)

    def groups(self): # TODO recurse?
        """Return all :class:`groups <Group>` in group."""
        return self.members(users=False)

    def members(self, groups=True, users=True): # TODO recurse?
        """Return all members in group (:class:`users <User>` or
        :class:`groups <Group>`)."""
        for ecuser in self.server.sa.GetUserListOfGroup(
                self._ecgroup.GroupID, MAPI_UNICODE):
            if ecuser.Username == 'SYSTEM':
                continue
            if users:
                try:
                    # TODO working around '@' duplication
                    username = '@'.join(ecuser.Username.split('@')[:2])
                    yield _user.User(username, self.server)
                    # TODO 'everyone', groups are included as users..
                except NotFoundError:
                    pass
            if groups:
                try:
                    yield Group(ecuser.Username, self.server)
                except NotFoundError:
                    pass

    @property
    def name(self):
        """Group name."""
        return self._name

    @name.setter
    def name(self, value):
        self._update(name=str(value))

    @property
    def email(self):
        """Group email address."""
        return self._ecgroup.Email

    @email.setter
    def email(self, value):
        self._update(email=str(value))

    @property
    def fullname(self): # currently identical to name (may change later?)
        return self._ecgroup.Fullname

    @fullname.setter
    def fullname(self, value):
        self._update(fullname=str(value))

    @property
    def hidden(self):
        """The group is hidden from the addressbook."""
        return bool(self._ecgroup.IsHidden)

    @hidden.setter
    def hidden(self, value):
        self._update(hidden=value)

    def send_as(self):
        """Return :class:`users <User>` in send-as list."""
        for u in self.server.sa.GetSendAsList(
                self._ecgroup.GroupID,MAPI_UNICODE):
            yield self.server.user(u.Username)

    def add_send_as(self, user):
        """Add :class:`user <User>` to send-as list.

        :param user: user to add
        """
        try:
            self.server.sa.AddSendAsUser(
                self._ecgroup.GroupID, user._ecuser.UserID)
        except MAPIErrorCollision:
            raise DuplicateError(
                "user '%s' already in send-as for group '%s'" %
                    (user.name, self.name))

    def remove_send_as(self, user):
        """Remove :class:`user <User>` from send-as list.

        :param user: user to remove
        """
        try:
            self.server.sa.DelSendAsUser(
                self._ecgroup.GroupID, user._ecuser.UserID)
        except MAPIErrorNotFound:
            raise NotFoundError(
                "no user '%s' in send-as for group '%s'" %
                    (user.name, self.name))

    # TODO: also does groups..
    def add_user(self, user):
        """Add :class:`user <User>` to group.

        :param user: user to add
        """
        if isinstance(user, Group):
            self.server.sa.AddGroupUser(
                self._ecgroup.GroupID, user._ecgroup.GroupID)
        else:
            try:
                self.server.sa.AddGroupUser(
                    self._ecgroup.GroupID, user._ecuser.UserID)
            except MAPIErrorCollision:
                raise DuplicateError(
                    "group '%s' already contains user '%s'" %
                        (self.name, user.name))

    def remove_user(self, user): # TODO also 'delete' for consistency?
        """Remove :class:`user <User>` from group.

        :param user: user to remove
        """
        try:
            self.server.sa.DeleteGroupUser(
                self._ecgroup.GroupID, user._ecuser.UserID)
        except MAPIErrorNotFound:
            raise NotFoundError(
                "group '%s' does not contain user '%s'" %
                    (self.name, user.name))

    def _update(self, **kwargs):
        # TODO: crashes server on certain characters...
        self._name = kwargs.get('name', self.name)
        fullname = kwargs.get('fullname', self.fullname)
        email = kwargs.get('email', self.email)
        hidden = kwargs.get('hidden', self.hidden)
        group = ECGROUP(self._name, fullname, email, int(hidden),
            self._ecgroup.GroupID)
        self.server.sa.SetGroup(group, MAPI_UNICODE)
        self._ecgroup = self.server.sa.GetGroup(
            self.server.sa.ResolveGroupName(self._name, MAPI_UNICODE),
            MAPI_UNICODE
        )

    def __eq__(self, u): # TODO check same server?
        if isinstance(u, Group):
            return self.groupid == u.groupid
        return False

    def __ne__(self, g):
        return not self == g

    def __contains__(self, u): # TODO subgroups
        return u in self.users()

    def __unicode__(self):
        return "Group('%s')" % self.name



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import codecs
import datetime
import email.parser as email_parser
import email.utils as email_utils
import functools
import os
import sys
import traceback

import pickle

import inetmapi
import icalmapi

from MAPI import (
    PT_MV_BINARY, PT_BOOLEAN, MSGFLAG_READ,
    CLEAR_READ_FLAG, PT_MV_UNICODE, PT_BINARY, PT_LONG,
    MAPI_CREATE, MAPI_MODIFY, MAPI_DEFERRED_ERRORS,
    ATTACH_BY_VALUE, ATTACH_EMBEDDED_MSG, STGM_WRITE, STGM_TRANSACTED,
    MAPI_UNICODE, MAPI_TO, MAPI_CC, MAPI_BCC, MAPI_E_NOT_FOUND,
    MAPI_E_NOT_ENOUGH_MEMORY, MAPI_ASSOCIATED,
    WrapCompressedRTFStream, MODRECIP_ADD, MODRECIP_REMOVE,
    RELOP_EQ
)

from MAPI.Util.codepage import GetCharsetByCP

from MAPI.Defs import (
    HrGetOneProp, CHANGE_PROP_TYPE
)

from MAPI.Struct import (
    SPropValue, MAPIErrorNotFound, MAPIErrorUnknownEntryid,
    MAPIErrorInterfaceNotSupported, MAPIErrorUnconfigured, MAPIErrorNoAccess,
    MAPIErrorNoRecipients, SPropertyRestriction
)

from MAPI.Tags import (
    PR_BODY, PR_DISPLAY_NAME_W, PR_MESSAGE_CLASS_W, PR_CHANGE_KEY,
    PR_CONTAINER_CLASS_W, PR_ENTRYID, PR_EC_HIERARCHYID, PR_HASATTACH,
    PR_SOURCE_KEY, PR_SUBJECT_W, PR_ATTACH_LONG_FILENAME_W, PR_INTERNET_CPID,
    PR_MESSAGE_SIZE, PR_BODY_W, PR_CREATION_TIME, PR_CLIENT_SUBMIT_TIME,
    PR_MESSAGE_DELIVERY_TIME, PR_LAST_MODIFICATION_TIME,
    PR_MESSAGE_FLAGS, PR_PARENT_ENTRYID, PR_IMPORTANCE,
    PR_ATTACH_NUM, PR_ATTACH_METHOD, PR_ATTACH_DATA_BIN,
    PR_TRANSPORT_MESSAGE_HEADERS, PR_ATTACH_DATA_OBJ, PR_ICON_INDEX,
    PR_EC_IMAP_EMAIL, PR_SENTMAIL_ENTRYID, PR_DELETE_AFTER_SUBMIT,
    PR_SENDER_ADDRTYPE_W, PR_SENDER_NAME_W, PR_SENDER_EMAIL_ADDRESS_W,
    PR_SENDER_ENTRYID, PR_SENT_REPRESENTING_ADDRTYPE_W,
    PR_SENT_REPRESENTING_NAME_W, PR_SENT_REPRESENTING_EMAIL_ADDRESS_W,
    PR_SENT_REPRESENTING_ENTRYID, PR_MESSAGE_RECIPIENTS,
    PR_MESSAGE_ATTACHMENTS, PR_RECIPIENT_TYPE, PR_ADDRTYPE_W,
    PR_EMAIL_ADDRESS_W, PR_SMTP_ADDRESS_W, PR_NULL, PR_HTML,
    PR_RTF_COMPRESSED, PR_SENDER_SEARCH_KEY,
    PR_SENT_REPRESENTING_SEARCH_KEY, PR_ATTACHMENT_FLAGS,
    PR_ATTACHMENT_HIDDEN, PR_ATTACHMENT_LINKID, PR_ATTACH_FLAGS,
    PR_NORMALIZED_SUBJECT_W, PR_INTERNET_MESSAGE_ID_W, PR_CONVERSATION_ID,
    PR_READ_RECEIPT_REQUESTED, PR_ORIGINATOR_DELIVERY_REPORT_REQUESTED,
    PR_REPLY_RECIPIENT_ENTRIES, PR_SENSITIVITY,
    PR_SEARCH_KEY, PR_LAST_VERB_EXECUTED, PR_LAST_VERB_EXECUTION_TIME,
    PR_ROWID, respOrganized
)

from MAPI.Tags import (
    IID_IAttachment, IID_IStream, IID_IMAPITable, IID_IMailUser, IID_IMessage,
)

from .pidlid import (
    PidLidAppointmentStateFlags, PidLidResponseStatus,
    PidLidAppointmentStartWhole, PidLidAppointmentEndWhole, PidLidFInvited
)

from .compat import (
    pickle_load as _pickle_load,
    pickle_loads as _pickle_loads,
    is_file as _is_file, benc as _benc, bdec as _bdec,
    default as _default,
)

from .defs import (
    NAMED_PROPS_ARCHIVER, NAMED_PROP_CATEGORY, ADDR_PROPS,
    PSETID_Archive, URGENCY, REV_URGENCY, ASF_MEETING,
)
from .errors import (
    Error, NotFoundError
)

from .attachment import Attachment
from .properties import Properties
from .recurrence import Occurrence
from .restriction import Restriction
from .meetingrequest import MeetingRequest, _copytags, _create_meetingrequest
from .distlist import DistList
from .address import Address
from .table import Table
from .contact import Contact
from .appointment import Appointment

try:
    from . import folder as _folder
except ImportError:  # pragma: no cover
    _folder = sys.modules[__package__+'.folder']
try:
    from . import user as _user
except ImportError:  # pragma: no cover
    _user = sys.modules[__package__ + '.user']
try:
    from . import utils as _utils
except ImportError:  # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']
from . import property_ as _prop

TESTING = False
if os.getenv('PYKO_TESTING'):  # env variable used in testset
    TESTING = True

class PersistentList(list):
    def __init__(self, mapiobj, proptag, *args, **kwargs):
        self.mapiobj = mapiobj
        self.proptag = proptag
        for attr in ('append', 'extend', 'insert', 'pop', 'remove', 'reverse', 'sort'):
            setattr(self, attr, self._autosave(getattr(self, attr)))
        list.__init__(self, *args, **kwargs)

    def _autosave(self, func):
        @functools.wraps(func)
        def _func(*args, **kwargs):
            ret = func(*args, **kwargs)
            data = [str(x) for x in self]
            self.mapiobj.SetProps([SPropValue(self.proptag, data)])
            _utils._save(self.mapiobj)
            return ret
        return _func

class Item(Properties, Contact, Appointment):
    """Item class

    Abstraction for basic items, such as emails, calendar appointments
    and contacts.

    Includes all functionality from :class:`Appointment`,
    :class:`Contact` and :class:`Properties`.
    """

    def __init__(self, folder=None, eml=None, ics=None, vcf=None, load=None,
                 loads=None, attachments=True, create=False, mapiobj=None,
                 entryid=None, content_flag=None, cache={}, save=True, read=None):
        self._eml = None
        self._architem = None
        self._folder = None
        self.mapiobj = mapiobj
        self._entryid = entryid
        self._content_flag = content_flag or 0
        self._cache = cache

        if folder:
            self._folder = folder
            self.store = folder.store
            self.server = folder.server

        if create:
            self.mapiobj = self.folder.mapiobj.CreateMessage(None, MAPI_ASSOCIATED if self.folder._content_flag & MAPI_ASSOCIATED else 0)
            self.store = self.folder.store
            self.server = server = self.store.server  # TODO

            if _is_file(eml):
                eml = eml.read()
            if _is_file(ics):
                ics = ics.read()
            if _is_file(vcf):
                vcf = vcf.read()

            self._eml = eml

            if eml is not None:
                # options for CreateMessage: 0 / MAPI_ASSOCIATED
                dopt = inetmapi.delivery_options()
                inetmapi.IMToMAPI(server.mapisession, self.folder.store.mapiobj, None, self.mapiobj, self._eml, dopt)
            elif ics is not None:
                icm = icalmapi.CreateICalToMapi(self.mapiobj, server.ab, False)
                # User object is not always available, when for example creating an item in the public folder
                user = self.user.mapiobj if self.user and self.user.mapiobj else None
                icm.ParseICal(ics, 'utf-8', 'UTC', user, 0)
                icm.GetItem(0, 0, self.mapiobj)
            elif vcf is not None:
                vcm = icalmapi.create_vcftomapi(self.mapiobj)
                vcm.parse_vcf(vcf)
                vcm.get_item(self.mapiobj)
            elif load is not None:
                self.load(load, attachments=attachments)
            elif loads is not None:
                self.loads(loads, attachments=attachments)
            else:
                try:
                    container_class = HrGetOneProp(self.folder.mapiobj, PR_CONTAINER_CLASS_W).Value
                except MAPIErrorNotFound:
                    self.mapiobj.SetProps(
                        [SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Note')])
                else:
                    # TODO just skip first 4 chars?
                    # TODO set all default props
                    if container_class == 'IPF.Contact':
                        self.mapiobj.SetProps(
                            [SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Contact')])
                    elif container_class == 'IPF.Appointment':
                        self.mapiobj.SetProps(
                            [SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Appointment')])
                        self.from_ = self.store.user
                        self[PidLidAppointmentStateFlags] = 0
                        # TODO move appt creation to appointment.py
                        self[PidLidResponseStatus] = 0
                        self[PidLidFInvited] = False

            if read is not None:
                self.read = read
            if save:
                _utils._save(self.mapiobj)

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if not self._mapiobj:
            self.mapiobj = _utils.openentry_raw(self.store.mapiobj, self._entryid, self._content_flag)
        return self._mapiobj

    @mapiobj.setter
    def mapiobj(self, mapiobj):
        self._mapiobj = mapiobj

    @property
    def _arch_item(self):
        # open archive store explicitly so we can handle otherwise silenced
        # errors (MAPI errors in mail bodies for example)
        if self._architem is None:
            if self.stubbed:
                ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
                PROP_STORE_ENTRYIDS = CHANGE_PROP_TYPE(ids[0], PT_MV_BINARY)
                PROP_ITEM_ENTRYIDS = CHANGE_PROP_TYPE(ids[1], PT_MV_BINARY)

                # support for multiple archives was a mistake, and is not and
                # _should not_ be used. so we just pick nr 0.
                arch_storeid = HrGetOneProp(self.mapiobj, PROP_STORE_ENTRYIDS).Value[0]
                item_entryid = HrGetOneProp(self.mapiobj, PROP_ITEM_ENTRYIDS).Value[0]

                try:
                    arch_store = self.server._store2(arch_storeid)
                except MAPIErrorUnconfigured:
                    raise Error(
                        'could not open archive store (check SSL settings?)')

                self._architem = arch_store.OpenEntry(item_entryid, None, 0)
            else:
                self._architem = self.mapiobj
        return self._architem

    @property
    def entryid(self):
        """Item entryid."""
        if self._entryid:
            return _benc(self._entryid)

        return _benc(self._get_fast(PR_ENTRYID, must_exist=True))

    @property
    def guid(self):
        return self.entryid[56:88]

    @property
    def hierarchyid(self):
        """Hierarchy (SQL) id."""
        return HrGetOneProp(self.mapiobj, PR_EC_HIERARCHYID).Value

    @property
    def sourcekey(self):
        """Item sourcekey."""
        # TODO more general caching solution
        if not hasattr(self, '_sourcekey'):
            self._sourcekey = _benc(self._get_fast(PR_SOURCE_KEY, must_exist=True))
        return self._sourcekey

    @property
    def searchkey(self):
        """Item searchkey."""
        # TODO more general caching solution
        if not hasattr(self, '_searchkey'):
            self._searchkey = _benc(HrGetOneProp(self.mapiobj, PR_SEARCH_KEY).Value)
        return self._searchkey

    @property
    def changekey(self):
        """Item changekey."""
        return _benc(self._get_fast(PR_CHANGE_KEY, must_exist=True))

    @property
    def subject(self):
        """Item subject."""
        return self._get_fast(PR_SUBJECT_W, _default(''))

    @subject.setter
    def subject(self, x):
        if x is None:
            prop = self.get_prop(PR_SUBJECT_W)
            if prop:
                self.delete(prop)
        else:
            self._set_fast(PR_SUBJECT_W, str(x))

    @property
    def name(self):
        """Item (display) name."""
        try:
            return self.prop(PR_DISPLAY_NAME_W).value
        except NotFoundError:
            return ''

    @name.setter
    def name(self, x):
        self.mapiobj.SetProps([SPropValue(PR_DISPLAY_NAME_W, str(x))])
        _utils._save(self.mapiobj)

    @property
    def normalized_subject(self):
        """Normalized item subject."""
        try:
            return self.prop(PR_NORMALIZED_SUBJECT_W).value
        except NotFoundError:
            return ''

    @property
    def body_preview(self):
        """Item body preview (plaintext, up to 255 characters)."""
        text = self._get_fast(PR_BODY_W, capped=True)
        if text:
            text = text[:255]
        return text

    @property
    def size(self):
        """Item storage size."""
        return self.prop(PR_MESSAGE_SIZE).value

    @property
    def message_class(self):
        """Item message class."""
        return self._get_fast(PR_MESSAGE_CLASS_W, must_exist=True)

    @message_class.setter
    def message_class(self, messageclass):
        # FIXME: Add all possible PR_MESSAGE_CLASS values
        """
        MAPI Message classes:
        * IPM.Note.SMIME.MultipartSigned - smime signed email
        * IMP.Note                       - normal email
        * IPM.Note.SMIME                 - smime encypted email
        * IPM.StickyNote                 - note
        * IPM.Appointment                - appointment
        * IPM.Task                       - task
        """
        self.mapiobj.SetProps([SPropValue(PR_MESSAGE_CLASS_W, str(messageclass))])
        _utils._save(self.mapiobj)

    @property
    def created(self):
        """Creation time."""
        return self._get_fast(PR_CREATION_TIME)

    @property
    def received(self):
        """Delivery time."""
        return self._get_fast(PR_MESSAGE_DELIVERY_TIME)

    @received.setter
    def received(self, value):
        self._cache.pop(PR_MESSAGE_DELIVERY_TIME, None)  # TODO generalize
        self.prop(PR_MESSAGE_DELIVERY_TIME, create=True).value = value

    @property
    def sent(self):
        """Client submit time."""
        try:
            return self.prop(PR_CLIENT_SUBMIT_TIME).value
        except NotFoundError:
            pass

    @property
    def last_modified(self):
        """Last modification time."""
        return self._get_fast(PR_LAST_MODIFICATION_TIME)

    @property
    def stubbed(self):
        """Item is stubbed by archiver."""
        # TODO caching folder.GetIDs..?
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_STUBBED = CHANGE_PROP_TYPE(ids[2], PT_BOOLEAN)
        try:
            # False means destubbed
            return HrGetOneProp(self.mapiobj, PROP_STUBBED).Value
        except MAPIErrorNotFound:
            return False

    # TODO draft status, how to determine? GetMessageStatus?

    @property
    def read(self):
        """Item is read."""
        return self.prop(PR_MESSAGE_FLAGS).value & MSGFLAG_READ > 0

    @read.setter
    def read(self, value):
        if value:
            self.mapiobj.SetReadFlag(0)
        else:
            self.mapiobj.SetReadFlag(CLEAR_READ_FLAG)

    @property
    def read_receipt(self):
        """A read receipt was requested."""
        return self.get(PR_READ_RECEIPT_REQUESTED, False)

    @read_receipt.setter
    def read_receipt(self, value):
        self.create_prop(PR_READ_RECEIPT_REQUESTED, value)

    @property
    def delivery_receipt(self):
        """A delivery receipt was requested."""
        return self.get(PR_ORIGINATOR_DELIVERY_REPORT_REQUESTED, False)

    @property
    def categories(self):
        """Item categories (user defined)."""
        proptag = self.mapiobj.GetIDsFromNames([NAMED_PROP_CATEGORY], MAPI_CREATE)[0]
        proptag = CHANGE_PROP_TYPE(proptag, PT_MV_UNICODE)
        try:
            value = self.prop(proptag).value
        except NotFoundError:
            value = []
        return PersistentList(self.mapiobj, proptag, value)

    @categories.setter
    def categories(self, value):
        proptag = self.mapiobj.GetIDsFromNames([NAMED_PROP_CATEGORY], MAPI_CREATE)[0]
        proptag = CHANGE_PROP_TYPE(proptag, PT_MV_UNICODE)
        data = [str(x) for x in value]
        self.mapiobj.SetProps([SPropValue(proptag, data)])
        _utils._save(self.mapiobj)

    @property
    def folder(self):
        """Item parent :class:`folder <Folder>`."""
        if self._folder:
            return self._folder
        try:
            return _folder.Folder(self.store, _benc(HrGetOneProp(self.mapiobj, PR_PARENT_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @property
    def urgency(self):
        """Item urgency (low, high or normal)."""
        try:
            return URGENCY[self[PR_IMPORTANCE]]
        except NotFoundError:
            return 'normal'

    @urgency.setter
    def urgency(self, value):
        self.create_prop(PR_IMPORTANCE, REV_URGENCY[value])

    @property
    def user(self):
        """Item :class:`owner <User>`."""
        return self.store.user

    @property
    def sensitivity(self):
        """Item sensitivity (normal, personal, private or confidential)."""
        try:
            return {
                0: 'normal',
                1: 'personal',
                2: 'private',
                3: 'confidential',
            }[self[PR_SENSITIVITY]]
        except NotFoundError:
            return 'normal'

    @property
    def private(self):
        """Should client hide item from other users."""
        try:
            return self.prop('common:34054').value
        except NotFoundError:
            return False

    @private.setter
    def private(self, value):
        return self.create_prop('common:34054', value, proptype=PT_BOOLEAN)

    @property
    def messageid(self):
        """Item internet message ID."""
        return self.get(PR_INTERNET_MESSAGE_ID_W)

    @property
    def conversationid(self):
        """Item conversation ID."""
        try:
            return _benc(self[PR_CONVERSATION_ID])
        except NotFoundError:
            pass

    def attachments(self, embedded=False, page_start=None, page_limit=None,
                    order=None):
        """Return all item :class:`attachments <Attachment>`

        :param embedded: include embedded attachments (default False)
        """
        mapiitem = self._arch_item
        table = Table(
            self.server,
            self.mapiobj,
            mapiitem.GetAttachmentTable(MAPI_DEFERRED_ERRORS),
            columns=[PR_ATTACH_NUM, PR_ATTACH_METHOD],
        )

        for row in table.rows():
            if(row[1].value == ATTACH_BY_VALUE or \
               (embedded and row[1].value == ATTACH_EMBEDDED_MSG)):
                yield Attachment(self, mapiitem, row[0].value)

    def attachment(self, entryid):
        # TODO can we use something existing for entryid, like PR_ENTRYID?
        # check exchange?
        # TODO performance, restriction on PR_RECORD_KEY part?
        for attachment in self.attachments(embedded=True):
            if attachment.entryid == entryid:
                return attachment
        else:
            raise NotFoundError("no attachment with entryid '%s'" % entryid)

    def create_attachment(self, name=None, data=None, filename=None, **kwargs):
        """Create an item attachment

        :param name: The attachment name
        :param data: Attachment data
        """

        if filename:
            data = open(filename, 'rb').read()
            name = os.path.basename(filename)

        (id_, attach) = self.mapiobj.CreateAttach(None, 0)
        name = str(name)
        props = [SPropValue(PR_ATTACH_LONG_FILENAME_W, name), SPropValue(PR_ATTACH_METHOD, ATTACH_BY_VALUE)]
        attach.SetProps(props)
        stream = attach.OpenProperty(PR_ATTACH_DATA_BIN, IID_IStream, STGM_WRITE | STGM_TRANSACTED, MAPI_MODIFY | MAPI_CREATE)
        stream.Write(data)
        stream.Commit(0)

        _utils._save(attach)
        _utils._save(self.mapiobj)  # TODO needed?

        att = Attachment(self, mapiitem=self.mapiobj, mapiobj=attach)

        for key, val in kwargs.items():
            setattr(att, key, val)

        return att

    @property
    def has_attachments(self):
        """Item has attachments."""
        return self.get(PR_HASATTACH, False)

    def header(self, name):
        """Item transport message header for given name

        :param name: Header name
        """
        return self.headers().get(name)

    def headers(self):
        """Item transport message headers."""
        try:
            message_headers = self.prop(PR_TRANSPORT_MESSAGE_HEADERS)
            headers = email_parser.BytesParser().parsebytes(
                message_headers.value, headersonly=True)
            return headers
        except NotFoundError:
            return {}

    @property
    def text(self):
        """Item plaintext representation."""
        try:
            return _utils.stream(self._arch_item, PR_BODY_W)
        except MAPIErrorNotFound:
            return ''

    @text.setter
    def text(self, x):
        self.create_prop(PR_BODY_W, str(x))

    @property
    def html(self):
        """Item HTML representation."""
        try:
            return _utils.stream(self._arch_item, PR_HTML)
        except MAPIErrorNotFound:
            return b''

    @html.setter
    def html(self, x):
        self.create_prop(PR_HTML, x)

    @property
    def html_utf8(self):
        """Item HTML representation (utf-8 encoded)."""
        if self.encoding == 'utf-8':
            return self.html
        else:
            html = codecs.decode(
                self.html, self.encoding or 'ascii', 'replace')
            # TODO set meta charset to utf-8!?
            return html.encode('utf-8')

    @property
    def rtf(self):
        """Item RTF representation."""
        try:
            return _utils.stream(self._arch_item, PR_RTF_COMPRESSED)
        except MAPIErrorNotFound:
            return b''

    @rtf.setter
    def rtf(self, x):
        stream = self._arch_item.OpenProperty(PR_RTF_COMPRESSED, IID_IStream, STGM_WRITE | STGM_TRANSACTED, MAPI_MODIFY | MAPI_CREATE)
        uncompressed = WrapCompressedRTFStream(stream, MAPI_MODIFY)
        uncompressed.Write(x)
        uncompressed.Commit(0)
        stream.Commit(0)

    @property
    def body_type(self):
        """Original body type: text, html, rtf or *None*."""
        tag = _prop.bestbody(self.mapiobj)
        if tag == PR_BODY_W:
            return 'text'
        elif tag == PR_HTML:
            return 'html'
        elif tag == PR_RTF_COMPRESSED:
            return 'rtf'

    def _generate_eml(self, received_date):
        sopt = inetmapi.sending_options()
        sopt.no_recipients_workaround = True
        sopt.add_received_date = received_date
        return inetmapi.IMToINet(self.store.server.mapisession, None, self.mapiobj, sopt)

    def eml(self, received_date=False, stored=True):
        """Convert the item to eml (RFC 2822) data

        :param received_date: add delivery date as received date
        :param stored: if available, use the stored version instead of
                       converting (default True)
        """
        if not stored:
            return self._generate_eml(received_date)

        if self._eml is None:
            try:
                self._eml = _utils.stream(self.mapiobj, PR_EC_IMAP_EMAIL)
            except MAPIErrorNotFound:
                self._eml = self._generate_eml(received_date)
        return self._eml

    def vcf(self):
        """Convert the item to vCard data."""
        vic = icalmapi.create_mapitovcf()
        vic.add_message(self.mapiobj)
        return vic.finalize()

    def ics(self, charset="UTF-8"):
        """Convert the item to iCal data."""
        mic = icalmapi.CreateMapiToICal(self.server.ab, charset)
        mic.AddMessage(self.mapiobj, "", 0)
        _, data = mic.Finalize(0)
        return data

    def _generate_reply_body(self):
        """Create a reply body"""
        # TODO(jelle): HTML formatted text support.
        body = "\r\n".join("> " + line for line in self.text.split('\r\n'))

        if self.sent and self.from_:
            body = "\r\nOn {} at {}, {} wrote:\r\n{}".format(
                self.sent.strftime('%d-%m-%Y'),
                self.sent.strftime('%H:%M'),
                self.from_.name,
                body)

        return body

    def _create_source_message_info(self, action):
        """Create source message info, value of the record, based on action
        type (reply, replyall, forward) and add 48byte entryid at the end.

        :param action: the reply action
        :type action: str
        :rtype: str
        """

        id_map = {
            'reply': '0501000066000000',
            'replyall': '0501000067000000',
            'forward': '0601000068000000',
        }

        try:
            return "01000E000C000000{0}0200000030000000{1}".format(
                id_map[action], self.entryid)
        except KeyError:
            return ""

    def reply(self, folder=None, all=False):
        """Create reply message

        :param folder: Folder to save reply (default in drafts)
        :param all: Reply to all (default False)
        """
        # TODO(jelle): support reply all.
        if not folder:
            folder = self.store.drafts
        # TODO(jelle): Remove multiple Re:'s, should only be one
        # Re: <subject> left
        subject = 'Re: {}'.format(self.subject)
        body = self._generate_reply_body()
        # TODO(jelle): pass an Address in to?
        item = folder.create_item(subject=subject, body=body,
            from_=self.store.user, to=self.from_.email,
            message_class=self.message_class)

        source_message_info = self._create_source_message_info(
            'reply').encode('ascii')
        item.create_prop('common:0x85CE', source_message_info, PT_BINARY)
        return item

    # TODO too many cases overlapping here
    def send(self, copy_to_sentmail=True, _basedate=None, cal_item=None):
        """Send item (mail or appointment as meeting request).

        :param copy_to_sentmail: Save copy in sentmail (default True)
        """
        item = self
        appt = None
        if self.message_class in (
            'IPM.Appointment',
            # exception message:
            'IPM.OLE.CLASS.{00061055-0000-0000-C000-000000000046}'
        ):
            if self.get(PidLidAppointmentStartWhole) is None or self.get(PidLidAppointmentEndWhole) is None:
                raise Error('appointment requires start and end date')

            appt = cal_item or self
            item = _create_meetingrequest(appt, self, basedate=_basedate)

        icon_index = {
            b'66': 261,  # reply
            b'67': 261,  # reply all
            b'68': 262,  # forward
        }
        # TODO(jelle): check if property exists?
        try:
            source_message = item.prop('common:0x85CE').value
            msgtype = source_message[24:26]
            orig = self.store.item(entryid=source_message[48:])
            orig.create_prop(PR_ICON_INDEX, icon_index[msgtype])
            orig.create_prop(PR_LAST_VERB_EXECUTED, int(msgtype, 16))
            orig.create_prop(PR_LAST_VERB_EXECUTION_TIME, datetime.datetime.now())
        except NotFoundError:
            pass

        props = []

        if copy_to_sentmail:
            props.append(SPropValue(PR_SENTMAIL_ENTRYID, _bdec(item.store.sentmail.entryid)))
        props.append(SPropValue(PR_DELETE_AFTER_SUBMIT, True))
        item.mapiobj.SetProps(props)

        try:
            item.mapiobj.SubmitMessage(0)
        except MAPIErrorNoRecipients:
            if self.message_class != 'IPM.Appointment':
                raise Error('cannot send item without recipients')

        if appt:
            appt[PidLidFInvited] = True

    @property
    def sender(self):
        """Item sender :class:`Address`."""
        addrprops = (PR_SENDER_ADDRTYPE_W, PR_SENDER_NAME_W, PR_SENDER_EMAIL_ADDRESS_W, PR_SENDER_ENTRYID, PR_SENDER_SEARCH_KEY)
        args = [self.get_prop(p).value if self.get_prop(p) else None for p in addrprops]
        return Address(self.server, *args)

    @sender.setter
    def sender(self, addr):
        pr_addrtype, pr_dispname, pr_email, pr_entryid = self._addr_props(addr)
        self.mapiobj.SetProps([
            # TODO pr_addrtype should be unicode already
            SPropValue(PR_SENDER_ADDRTYPE_W, str(pr_addrtype)),
            SPropValue(PR_SENDER_NAME_W, pr_dispname),
            SPropValue(PR_SENDER_EMAIL_ADDRESS_W, pr_email),
            SPropValue(PR_SENDER_ENTRYID, pr_entryid),
        ])
        _utils._save(self.mapiobj)

    @property
    def from_(self):
        """Item from :class:`Address` (not necessarily the sender)."""
        addrprops = (
            PR_SENT_REPRESENTING_ADDRTYPE_W,
            PR_SENT_REPRESENTING_NAME_W,
            PR_SENT_REPRESENTING_EMAIL_ADDRESS_W,
            PR_SENT_REPRESENTING_ENTRYID,
            PR_SENT_REPRESENTING_SEARCH_KEY
        )
        args = [self.get_prop(p).value if self.get_prop(p) else None for p in addrprops]
        return Address(self.server, *args)

    @from_.setter
    def from_(self, addr):
        pr_addrtype, pr_dispname, pr_email, pr_entryid = self._addr_props(addr)
        self.mapiobj.SetProps([
            # TODO pr_addrtype should be unicode already
            SPropValue(PR_SENT_REPRESENTING_ADDRTYPE_W, str(pr_addrtype)),
            SPropValue(PR_SENT_REPRESENTING_NAME_W, pr_dispname),
            SPropValue(PR_SENT_REPRESENTING_EMAIL_ADDRESS_W, pr_email),
            SPropValue(PR_SENT_REPRESENTING_ENTRYID, pr_entryid),
        ])

        # TODO Hack around missing sender
        if not self.sender.email:
            self.sender = addr

        _utils._save(self.mapiobj)

    def table(self, name, restriction=None, order=None, columns=None):
        return Table(
            self.server,
            self.mapiobj,
            self.mapiobj.OpenProperty(name, IID_IMAPITable, MAPI_UNICODE, 0),
            name,
            restriction=restriction,
            order=order,
            columns=columns,
        )

    def tables(self):
        yield self.table(PR_MESSAGE_RECIPIENTS)
        yield self.table(PR_MESSAGE_ATTACHMENTS)

    def recipients(self, _type=None):
        """Return all item recipient :class:`addresses <Address>`."""
        for row in self.table(PR_MESSAGE_RECIPIENTS):
            row2 = dict([(x.proptag, x) for x in row])
            if (not _type or (PR_RECIPIENT_TYPE in row2 and \
                row2[PR_RECIPIENT_TYPE].value == _type)):
                args = [row2[p].value if p in row2 else None
                    for p in (
                        PR_ADDRTYPE_W,
                        PR_DISPLAY_NAME_W,
                        PR_EMAIL_ADDRESS_W,
                        PR_ENTRYID,
                        PR_SEARCH_KEY)
                ]
                yield Address(self.server, *args, props=row)

    @property
    def to(self):
        """Return all item TO recipient :class:`addresses <Address>`."""
        return self.recipients(_type=MAPI_TO)

    @property
    def cc(self):
        """Return all item CC recipient :class:`addresses <Address>`."""
        return self.recipients(_type=MAPI_CC)

    @property
    def bcc(self):
        """Return all item BCC recipient :class:`addresses <Address>`."""
        return self.recipients(_type=MAPI_BCC)

    @property
    def replyto(self):
        """Return all :class:`Addresses <Address>` where reply should be
        sent."""
        try:
            entries = self[PR_REPLY_RECIPIENT_ENTRIES]
        except NotFoundError:
            return

        # specs say addressbook eids, but they are one-offs?
        # TODO use ECParseOneOff
        count = _utils.unpack_long(entries, 0)
        pos = 8
        for _ in range(count):
            size = _utils.unpack_long(entries, pos)
            content = entries[pos + 4 + 24:pos + 4 + 24 + size]

            start = sos = 0
            info = []
            while True:
                if content[sos:sos + 2] == b'\x00\x00':
                    info.append(content[start:sos].decode('utf-16-le'))
                    if len(info) == 3:
                        yield Address(self.server, info[1], info[0], info[2])
                        break
                    start = sos + 2
                sos += 2

            pos += 4 + size

    @property
    def is_meetingrequest(self):
        """Is the item a meeting request."""
        return self.message_class.startswith('IPM.Schedule.Meeting.')

    @property
    def meetingrequest(self):
        """:class:`Meeting request <MeetingRequest>` object."""
        return MeetingRequest(self)

    @property
    def is_distlist(self):
        """Is the item a distribution list."""
        return self.message_class == 'IPM.DistList'

    @property
    def distlist(self):
        """:class:`Distribution List <DistList>` object."""
        return DistList(self)

    def _addr_props(self, addr):
        if isinstance(addr, _user.User):
            pr_addrtype = 'ZARAFA'
            pr_dispname = addr.name
            pr_email = addr.email
            pr_entryid = _bdec(addr.userid)
        elif isinstance(addr, Address):
            pr_addrtype = addr.addrtype
            pr_dispname = addr.name
            pr_email = addr.email
            pr_entryid = addr.entryid
        else:
            addr = str(addr)
            pr_addrtype = 'SMTP'
            pr_dispname, pr_email = email_utils.parseaddr(addr)
            pr_dispname = pr_dispname or addr
            pr_entryid = self.server.ab.CreateOneOff(
                pr_dispname, 'SMTP', str(pr_email), MAPI_UNICODE)
        return pr_addrtype, pr_dispname, pr_email, pr_entryid

    def _recipients(self, reciptype, addrs):
        """Sets the recipient table according to the type and given addresses.

        Since the recipienttable contains to, cc and bcc addresses, and item.to
        = [] should only set the to recipients (not append), the to recipients
        are first removed and then added again. This approach works but seems
        suboptimal, since it queries the MAPI_TO items, removes them and adds
        the new entries while leaving the other recipients intact.

        :param reciptype: MAPI_TO, MAPI_CC or MAPI_BCC
        :param addrs: list of addresses to set.
        """

        restriction = Restriction(SPropertyRestriction(RELOP_EQ,
            PR_RECIPIENT_TYPE, SPropValue(PR_RECIPIENT_TYPE, reciptype)))
        table = self.table(PR_MESSAGE_RECIPIENTS, columns=[PR_ROWID],
            restriction=restriction)
        rows = [[SPropValue(PR_ROWID, row[PR_ROWID])]
            for row in table.dict_rows()]
        if rows:
            self.mapiobj.ModifyRecipients(MODRECIP_REMOVE, rows)

        if isinstance(addrs, str):
            addrs = [x.strip() for x in str(addrs).split(';')]
        elif isinstance(addrs, _user.User):
            addrs = [addrs]
        names = []
        for addr in addrs:
            pr_addrtype, pr_dispname, pr_email, pr_entryid = \
                self._addr_props(addr)

            names.append([
                SPropValue(PR_RECIPIENT_TYPE, reciptype),
                SPropValue(PR_DISPLAY_NAME_W, pr_dispname),
                SPropValue(PR_ADDRTYPE_W, str(pr_addrtype)),
                SPropValue(PR_EMAIL_ADDRESS_W, str(pr_email)),
                SPropValue(PR_ENTRYID, pr_entryid),
            ])
        self.mapiobj.ModifyRecipients(MODRECIP_ADD, names)

        if self.message_class == 'IPM.Appointment':
            self[PidLidAppointmentStateFlags] = ASF_MEETING
            self[PidLidResponseStatus] = respOrganized  # TODO delegation?

        _utils._save(self.mapiobj)  # TODO needed?

    @to.setter
    def to(self, addrs):
        self._recipients(MAPI_TO, addrs)

    @cc.setter
    def cc(self, addrs):
        self._recipients(MAPI_CC, addrs)

    @bcc.setter
    def bcc(self, addrs):
        self._recipients(MAPI_BCC, addrs)

    def delete(self, objects):
        """Delete properties or attachments from item.

        :param objects: The object(s) to delete
        """
        objects = _utils.arg_objects(objects, (Attachment, _prop.Property,
            Occurrence), 'Item.delete')
        # TODO embedded items?

        attach_ids = [item.number for item in objects
            if isinstance(item, Attachment)]
        proptags = [item.proptag for item in objects
            if isinstance(item, _prop.Property)]
        occs = [item for item in objects
            if isinstance(item, Occurrence)]
        for occ in occs:
            self.recurrence._delete_exception(occ.start, self,
                _copytags(self.mapiobj))
        if proptags:
            self.mapiobj.DeleteProps(proptags)
        for attach_id in attach_ids:
            self._arch_item.DeleteAttach(attach_id, 0, None, 0)

        # TODO: refresh the mapiobj since PR_ATTACH_NUM is updated when opening
        # a message? PR_HASATTACH is also updated by the server.
        _utils._save(self.mapiobj)

    def match(self, restriction):
        """Does the given restriction match with the item.

        :param restriction: :class:`Restriction <Restriction>`
        """
        return restriction.match(self)

    def _convert_to_smtp(self, props, tag_data):
        if not hasattr(self.server, '_smtp_cache'): # TODO speed hack, discuss
            self.server._smtp_cache = {}
        for addrtype, email, entryid, name, searchkey in ADDR_PROPS:
            if (addrtype not in tag_data or \
                entryid not in tag_data or \
                name not in tag_data):
                continue
            # TODO MAPIPDL==distlist.. can we just dump this?
            if tag_data[addrtype][1] in ('SMTP', 'MAPIPDL'):
                continue
            eid = tag_data[entryid][1]
            if eid in self.server._smtp_cache:
                email_addr = self.server._smtp_cache[eid]
            else:
                try:
                    mailuser = self.server.ab.OpenEntry(eid, IID_IMailUser, 0)
                    email_addr = HrGetOneProp(mailuser,
                        PR_SMTP_ADDRESS_W).Value
                except MAPIErrorUnknownEntryid:
                    # TODO corrupt data? keep going but log problem
                    continue
                except MAPIErrorNotFound:
                    # TODO deleted user, or no email address?
                    # or user with multiple entryids..heh?
                    continue
                except MAPIErrorInterfaceNotSupported: # TODO ZARAFA group?
                    continue
                self.server._smtp_cache[eid] = email_addr
            tag_data[addrtype][1] = 'SMTP'
            if email in tag_data:
                tag_data[email][1] = email_addr
            else:
                props.append([email, email_addr, None])
            tag_data[entryid][1] = self.server.ab.CreateOneOff(
                tag_data[name][1], 'SMTP', email_addr, MAPI_UNICODE)
            key = b'SMTP:' + email_addr.upper().encode('ascii')
            # TODO probably need to create, also email
            if searchkey in tag_data:
                tag_data[searchkey][1] = key
            else:
                props.append([searchkey, key, None])

    def _dump(self, attachments=True, archiver=True, skip_broken=False,
            _main_item=None):
        _main_item = _main_item or self
        log = self.server.log

        # props
        props = []
        tag_data = {}
        bestbody = _prop.bestbody(self.mapiobj)
        for prop in self.props():
            if bestbody != PR_NULL and prop.proptag in (PR_BODY_W, PR_HTML, PR_RTF_COMPRESSED) and prop.proptag != bestbody:
                continue
            if prop.named:  # named prop: prop.id_ system dependent...
                data = [
                    prop.proptag,
                    prop.mapiobj.Value,
                    self.mapiobj.GetNamesFromIDs([prop.proptag], None, 0)[0]
                ]
                if not archiver and data[2].guid == PSETID_Archive:
                    continue
            else:
                data = [prop.proptag, prop.mapiobj.Value, None]
            props.append(data)
            tag_data[prop.proptag] = data
        self._convert_to_smtp(props, tag_data)

        # recipients
        recs = []
        for row in self.table(PR_MESSAGE_RECIPIENTS):
            rprops = []
            tag_data = {}
            for prop in row:
                data = [prop.proptag, prop.mapiobj.Value, None]
                rprops.append(data)
                tag_data[prop.proptag] = data
            recs.append(rprops)
            self._convert_to_smtp(rprops, tag_data)

        # attachments
        atts = []
        # TODO optimize by looking at PR_MESSAGE_FLAGS?
        # TODO should we use GetAttachmentTable?
        for row in self.table(PR_MESSAGE_ATTACHMENTS).dict_rows():
            try:
                if TESTING and os.getenv('PYKO_TEST_KEY_ERROR'):
                    raise KeyError()
                num = row[PR_ATTACH_NUM]
                method = row.get(PR_ATTACH_METHOD, ATTACH_BY_VALUE)
                att = self.mapiobj.OpenAttach(num, IID_IAttachment, 0)
                if method == ATTACH_EMBEDDED_MSG:
                    try:
                        msg = att.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_DEFERRED_ERRORS | MAPI_MODIFY)
                    except MAPIErrorNoAccess:
                        # TODO the following may fail for embedded items in
                        # certain public stores, while the above does work
                        # (opening read-only doesn't work, but read-write
                        # works! wut!?)
                        msg = att.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_DEFERRED_ERRORS)
                    item = Item(mapiobj=msg)
                    item.server = self.server  # TODO
                    data = item._dump(_main_item=_main_item)  # recursion
                    atts.append(([[a, b, None] for a, b in row.items()], data))
                elif method == ATTACH_BY_VALUE and attachments:
                    try:
                        if TESTING and os.getenv('PYKO_TEST_NOT_FOUND'):
                            raise MAPIErrorNotFound()
                        data = _utils.stream(att, PR_ATTACH_DATA_BIN)
                    except MAPIErrorNotFound:
                        log.warning("no data found for attachment of item \
with entryid %s", _main_item.entryid)
                        data = ''
                    atts.append(([[a, b, None] for a, b in row.items()], data))
            except Exception as e:  # TODO generalize so usable in more places
                log.error('could not serialize attachment for item with \
entryid %s', _main_item.entryid)
                if skip_broken:
                    log.error(traceback.format_exc())
                    service = self.server.service
                    if service and service.stats:
                        service.stats['errors'] += 1
                else:
                    raise

        return {
            b'props': props,
            b'recipients': recs,
            b'attachments': atts,
        }

    # TODO f
    def dump(self, f, attachments=True, archiver=True):
        """Serialize item into open file.

        :param f: Open file
        """
        item = Item(mapiobj=self._arch_item)  # TODO make configurable?
        item.server = self.server
        pickle.dump(item._dump(attachments=attachments, archiver=archiver), f, protocol=2)

    def dumps(self, attachments=True, archiver=True, skip_broken=False):
        """Serialize item."""
        item = Item(mapiobj=self._arch_item)  # TODO make configurable?
        item.server = self.server
        return pickle.dumps(item._dump(attachments=attachments,
            archiver=archiver, skip_broken=skip_broken), protocol=2)

    def _load(self, d, attachments):
        # props
        props = []
        for proptag, value, nameid in d[b'props']:
            if nameid is not None:
                proptag = self.mapiobj.GetIDsFromNames(
                    [nameid], MAPI_CREATE)[0] | (proptag & 0xffff)
            if ((proptag >> 16) not in \
                (PR_BODY >> 16, PR_HTML >> 16, PR_RTF_COMPRESSED >> 16) or \
                value not in (MAPI_E_NOT_FOUND, MAPI_E_NOT_ENOUGH_MEMORY)):
                # TODO do we need to backup these errors
                props.append(SPropValue(proptag, value))
        self.mapiobj.SetProps(props)

        # recipients
        recipients = [[SPropValue(proptag_, value) for (proptag_, value, nameid) in row] for row in d[b'recipients']]
        self.mapiobj.ModifyRecipients(0, recipients)

        # attachments
        for props, data in d[b'attachments']:
            if attachments or isinstance(data, dict):
                props = [SPropValue(proptag_, value)
                    for (proptag_, value, nameid) in props]
                (id_, attach) = self.mapiobj.CreateAttach(None, 0)
                attach.SetProps(props)
                if isinstance(data, dict):  # embedded message
                    msg = attach.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY)
                    item = Item(mapiobj=msg)
                    item._load(data, attachments)  # recursion
                else:
                    stream = attach.OpenProperty(PR_ATTACH_DATA_BIN, IID_IStream, STGM_WRITE | STGM_TRANSACTED, MAPI_MODIFY | MAPI_CREATE)
                    stream.Write(data)
                    stream.Commit(0)
                _utils._save(attach)

        _utils._save(self.mapiobj) # TODO needed?

    # TODO f
    def load(self, f, attachments=True):
        """Deserialize item from open file

        :param f: Open file
        """
        self._load(_pickle_load(f), attachments)

    # TODO s
    def loads(self, s, attachments=True):
        """Deserialize item from data

        :param s: Serialized data
        """
        self._load(_pickle_loads(s), attachments)

    def create_item(self, message_flags=None, hidden=False, **kwargs):
        """Create embedded :class:`item <Item>`."""

        (id_, attach) = self.mapiobj.CreateAttach(None, 0)

        attach.SetProps([
            SPropValue(PR_ATTACH_METHOD, ATTACH_EMBEDDED_MSG),
            SPropValue(PR_ATTACHMENT_FLAGS, 2),
            SPropValue(PR_ATTACHMENT_HIDDEN, hidden),
            SPropValue(PR_ATTACHMENT_LINKID, 0),
            SPropValue(PR_ATTACH_FLAGS, 0),
        ])
        if kwargs.get('subject'):
            attach.SetProps([
                SPropValue(PR_DISPLAY_NAME_W, str(kwargs.get('subject'))),
            ])

        msg = attach.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY)
        if message_flags is not None:
            msg.SetProps([SPropValue(PR_MESSAGE_FLAGS, message_flags)])

        item = Item(mapiobj=msg)
        item.server = self.server
        item.store = self.store
        item._attobj = attach

        for key, val in kwargs.items():
            setattr(item, key, val)

        _utils._save(msg)
        _utils._save(attach)
        _utils._save(self.mapiobj)  # TODO needed?

        return item

    def items(self, recurse=True):
        """Return all embedded :class:`items <Item>`."""
        # TODO should we use GetAttachmentTable?
        for row in self.table(PR_MESSAGE_ATTACHMENTS).dict_rows():
            num = row[PR_ATTACH_NUM]
            method = row.get(PR_ATTACH_METHOD, ATTACH_BY_VALUE)
            if method == ATTACH_EMBEDDED_MSG:
                att = self.mapiobj.OpenAttach(num, IID_IAttachment, 0)
                msg = att.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage, 0, MAPI_MODIFY | MAPI_DEFERRED_ERRORS)

                item = Item(mapiobj=msg)
                item._attobj = att  # TODO
                item.server = self.server  # TODO
                # KC-1390, item should always have item.store
                if hasattr(self, 'store'):
                    item.store = self.store
                yield item

                if recurse:
                    for subitem in item.items():
                        yield subitem

    @property
    def primary_item(self):
        """Primary item (for archived :class:`items <Item>`)."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_REF_ITEM_ENTRYID = CHANGE_PROP_TYPE(ids[4], PT_BINARY)
        entryid = HrGetOneProp(self.mapiobj, PROP_REF_ITEM_ENTRYID).Value

        try:
            return self.folder.primary_store.item(entryid=_benc(entryid))
        except NotFoundError:
            pass

    def restore(self):
        """Restore item (for archived :class:`items <Item>`)."""
        # TODO cache folder.GetIDs..?
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_STUBBED = CHANGE_PROP_TYPE(ids[2], PT_BOOLEAN)
        PROP_REF_STORE_ENTRYID = CHANGE_PROP_TYPE(ids[3], PT_BINARY)
        PROP_REF_ITEM_ENTRYID = CHANGE_PROP_TYPE(ids[4], PT_BINARY)
        PROP_REF_PREV_ENTRYID = CHANGE_PROP_TYPE(ids[5], PT_BINARY)
        PROP_FLAGS = CHANGE_PROP_TYPE(ids[6], PT_LONG)

        # get/create primary item
        primary_item = self.primary_item
        if not primary_item:
            mapiobj = self.folder.primary_folder.mapiobj.CreateMessage(None, 0)
            new = True
        else:
            mapiobj = primary_item.mapiobj
            new = False

        if not new and not primary_item.stubbed:
            return
        # cleanup primary item
        mapiobj.DeleteProps([PROP_STUBBED, PR_ICON_INDEX])
        at = Table(
            self.server,
            self.mapiobj,
            mapiobj.GetAttachmentTable(0),
            columns=[PR_ATTACH_NUM],
        )
        for row in at.rows():
            mapiobj.DeleteAttach(row[0].value, 0, None, 0)

        # copy contents into it
        exclude_props = [
            PROP_REF_STORE_ENTRYID,
            PROP_REF_ITEM_ENTRYID,
            PROP_REF_PREV_ENTRYID,
            PROP_FLAGS
        ]
        self.mapiobj.CopyTo(None, exclude_props, 0, None, IID_IMessage, mapiobj, 0)

        _utils._save(mapiobj)

        # update backref
        if new:
            entryid = HrGetOneProp(mapiobj, PR_ENTRYID).Value
            self.mapiobj.SetProps([SPropValue(PROP_REF_ITEM_ENTRYID, entryid)])
            _utils._save(self.mapiobj)

    def copy(self, folder, _delete=False):
        """Copy item to folder (returning copied item).

        :param folder: Target folder
        """
        mapiobj = folder.mapiobj.CreateMessage(None, 0)
        self.mapiobj.CopyTo([], [], 0, None, IID_IMessage, mapiobj, 0)
        _utils._save(mapiobj)
        if _delete:
            self.folder.delete(self)
        item = Item(mapiobj=mapiobj)
        item.store = self.store  # TODO
        item.server = self.server
        return item

    def move(self, folder):
        """Move item to folder (returning moved item).

        :param folder: Target folder
        """
        return self.copy(folder, _delete=True)

    @property
    def codepage(self):
        """Windows codepage for item."""
        return self.get(PR_INTERNET_CPID)

    @codepage.setter
    def codepage(self, value):
        self[PR_INTERNET_CPID] = value

    @property
    def encoding(self):
        """HTML body encoding."""
        if self.codepage is not None:
            try:
                return GetCharsetByCP(self.codepage)
            except Exception:
                raise NotFoundError(
                    'no encoding for codepage: %d' % self.codepage)

    @property
    def type_(self):
        if (self.message_class is None or \
            self.message_class.startswith('IPM.Note') or \
            self.message_class.startswith('IPM.Schedule')):
            return 'mail'
        elif self.message_class == 'IPM.Appointment':
            return 'appointment'
        elif self.message_class == 'IPM.Contact':
            return 'contact'
        elif self.message_class == 'IPM.DistList':
            return 'distlist'
        else:
            return 'item'

    def __eq__(self, i):  # TODO check same store?
        if isinstance(i, Item):
            return self.entryid == i.entryid
        return False

    def __ne__(self, i):
        return not self == i

    def __iter__(self):
        return self.props()

    def __unicode__(self):
        return 'Item(%s)' % self.subject



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

# two-way conversion of timezones between MAPI and python
#
# note that pyko always outputs/expects naive system-local datetimes.
# (except for recurrence.{start,end}, which are timezone-local dates)
# this may have been a mistake, but OTOH, it does seem nice for interactive
# use. anyway this is not easy to change now. we intend to add a global
# flag to python-kopano, to make all datetimes timezone-aware.

import calendar
import datetime
import struct

from dateutil.relativedelta import (
    relativedelta, MO, TU, TH, FR, WE, SA, SU
)
import pytz
import tzlocal

RRULE_WEEKDAYS = {0: SU, 1: MO, 2: TU, 3: WE, 4: TH, 5: FR, 6: SA}
UTC = pytz.utc
LOCAL = tzlocal.get_localzone()
TZFMT = '<lll H HHHHHHHH H HHHHHHHH'

# convert MAPI timezone struct to datetime-compatible tzinfo class

class MAPITimezone(datetime.tzinfo):
    """Python datetime compatible MAPI timezone class.

    With this class, we can use Python builtin functionality to
    perform conversion between timezones, which avoids potential
    subtle/tricky issues.
    """

    def __init__(self, tzdata):
        # TODO more thoroughly check specs (MS-OXOCAL)
        self.timezone, \
        _, \
        self.timezonedst, \
        _, _, \
        self.dstendmonth, \
        self.dstendweek, \
        self.dstendday, \
        self.dstendhour, \
        _, _, _, _, _, \
        self.dststartmonth, \
        self.dststartweek, \
        self.dststartday, \
        self.dststarthour, \
        _, _, _ = \
            struct.unpack(TZFMT, tzdata)

    def _date(self, dt, dstmonth, dstweek, dstday, dsthour):
        d = datetime.datetime(dt.year, dstmonth, 1, dsthour)
        if dstday == 5: # last weekday of month
            d += relativedelta(
                months=1, days=-1, weekday=RRULE_WEEKDAYS[dstweek](-1))
        else:
            d += relativedelta(
                weekday=RRULE_WEEKDAYS[dstweek](dstday))
        return d

    def dst(self, dt):
        if self.dststartmonth == 0: # no DST
            return datetime.timedelta(0)

        start = self._date(dt, self.dststartmonth, self.dststartweek,
            self.dststartday, self.dststarthour)
        end = self._date(dt, self.dstendmonth, self.dstendweek,
            self.dstendday, self.dstendhour)

        # Can't compare naive to aware objects, so strip the timezone
        # from dt first.
        # TODO end < start case!
        dt = dt.replace(tzinfo=None)

        if ((start < end and start < dt < end) or \
            (start > end and not end < dt < start)):
            return datetime.timedelta(minutes=-self.timezone)
        else:
            return datetime.timedelta(minutes=0)

    def utcoffset(self, dt):
        return datetime.timedelta(minutes=-self.timezone) + self.dst(dt)

    def tzname(self, dt):
        return 'MAPITimezone(%s)' % self.utcoffset(dt)

    def __repr__(self):
        return 'MAPITimezone(<%d,dst:%d-%d>)' % (self.timezone, self.dststartmonth, self.dstendmonth)

# convert Olson timezone name to MAPI timezone struct

def _timezone_struct(name):
    # this is a bit painful. we need to generate a MAPI timezone structure
    # (rule) from an 'olson' name. on the one hand the olson db (as in pytz)
    # does not expose the actual DST rules, if any, and just expands all
    # transitions. this means we have to do some guess work.
    # on the other hand, the MAPI struct/rule is quite limited (transition
    # always on "Nth weekday in month", and not historical). ideally of
    # course we move away from the MAPI struct altogether, and just use the
    # olson name.. but that would probably break other clients at this point.

    tz = pytz.timezone(name)

    now = datetime.datetime.now()
    year = now.year

    yearstart = datetime.datetime(year,1,1)

    runningdst = tz.dst(yearstart)

    UTCBIAS = tz.utcoffset(yearstart)
    DSTBIAS = None
    DSTSTART = None
    DSTEND = None

    ZERO = datetime.timedelta(0)

    days = 366 if calendar.isleap(year) else 365

    for day in range(days): # TODO find matching transitions
        dt = datetime.datetime(year,1,1) + datetime.timedelta(days=day)

        datedst = tz.dst(dt)

        if runningdst == ZERO and datedst != ZERO:
            runningdst = datedst
            UTCBIAS = tz.utcoffset(dt) - datedst
            DSTBIAS = datedst
            DSTSTART = dt

        elif runningdst != ZERO and datedst == ZERO:
            runningdst = datedst
            DSTEND = dt

    utcbias = -UTCBIAS.seconds//60

    if DSTBIAS:
        dstbias = -DSTBIAS.seconds//60
        startmonth = DSTSTART.month
        endmonth = DSTEND.month
        # TODO don't assume start/end are same weekday and last-in-month
        weekday = DSTSTART.weekday()

        return struct.pack(TZFMT, utcbias, 0, dstbias, 0, 0, endmonth,
            weekday, 5, 0, 0, 0, 0, 0, 0, startmonth, weekday, 5, 0, 0, 0, 0)
    else:
        return struct.pack(TZFMT, utcbias, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0)

# conversion helpers

def _to_naive(date):
    if date.tzinfo is not None:
        date = date.replace(tzinfo=None)
    return date

def _to_tz(date, tzinfo):
    if hasattr(tzinfo, 'localize'):
        return tzinfo.localize(_to_naive(date))
    else:
        return date.replace(tzinfo=tzinfo)

def _from_utc(date, tzinfo):
    return UTC.localize(_to_naive(date)).astimezone(tzinfo).replace(tzinfo=None)

def _to_utc(date, tzinfo): # TODO local??
    return _to_tz(date, tzinfo).astimezone(UTC).replace(tzinfo=None)

def _tz2(date, tz1, tz2):
    return _to_tz(date, tz1).astimezone(tz2).replace(tzinfo=None)



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-or-later

from contextlib import closing
import os
import sys
import time

import kopano
from kopano.log import logger

import bsddb3 as bsddb

CONFIG = {
    "autorespond_cc": kopano.Config.boolean(default=False),
    "autorespond_bcc": kopano.Config.boolean(default=False),
    "autorespond_norecip": kopano.Config.boolean(default=False),
    "timelimit": kopano.Config.integer(default=24 * 3600),
    "senddb": kopano.Config.string(default="/var/db/kopano/autorespond.db"),
    "copy_to_sentmail": kopano.Config.boolean(default=True)
}

user_blacklist = ["mailer-daemon", "postmaster", "root"]


def in_blacklist(log, from_, to):
    short_from = from_.split("@")[0]
    short_to = to.split("@")[0]

    if short_from in user_blacklist or short_to in user_blacklist:
        log.info("From or To is blacklisted (from: %s, to: %s)", from_, to)
        return True

    return False


def send_ooo(server, username, msg, copy_to_sentmail):
    outbox = server.user(username).outbox
    item = outbox.create_item(eml=msg)
    item.send(copy_to_sentmail=copy_to_sentmail)


def check_time(senddb, timelimit, username, to):
    with closing(bsddb.btopen(senddb, 'c')) as db:
        key = username + ":" + to
        key = key.encode('utf-8')
        if key in db:
            timestamp = int(db[key])
            if timestamp + timelimit > int(time.time()):
                return True
    return False


def add_time(senddb, username, to):
    with closing(bsddb.btopen(senddb, 'c')) as db:
        key = username + ":" + to
        key = key.encode('utf-8')
        db[key] = str(int(time.time()))


def main():
    parser = kopano.parser("CSKQ", usage="Usage: %prog [options] from to subject username msgfile")
    options, args = parser.parse_args()

    if len(args) != 5:
        print("Invalid arguments, you have to supply the arguments: from, to, subject, username, and msgfile", file=sys.stderr)
        sys.exit(1)

    config_dict = kopano.CONFIG
    config_dict.update(CONFIG)
    config = kopano.Config(config_dict, options=options, service="autorespond")
    log = logger("autorespond", options=options, config=config)
    server = kopano.server(options=options, config=config)

    (from_, to, subject, username, msg_file) = args
    (to_me, bcc_me, cc_me) = (
        os.getenv("MESSAGE_TO_ME"), os.getenv("MESSAGE_BCC_ME"),
        os.getenv("MESSAGE_CC_ME"),
    )

    try:
        fh = open(msg_file, "rb")
        msg = fh.read()
        fh.close()
    except:
        log.info("Could not open msg file: %s", msg_file)
        sys.exit(1)

    if not(config["autorespond_norecip"] or
           (config["autorespond_bcc"] and bcc_me) or
           (config["autorespond_cc"] and cc_me) or to_me):
        log.debug("Response not send due to configuration")
    elif not(from_ and to and username and msg and msg_file):
        log.info("One of the input arguments was empty (from: %s, to: %s, username: %s, msg: %s)",
                 from_, to, username, msg)
    elif from_ == to:
        log.info("Loop detected, from == to (%s)", from_)
    elif in_blacklist(log, from_, to):
        log.info("From or to is in blacklist (from: %s, to: %s)", from_, to)
    elif check_time(config["senddb"], config["timelimit"], username, to):
        log.info("Ignoring since we already sent OOO message within timelimit (username: %s, to: %s, timelimit: %d)",
                 username, to, config["timelimit"])
    else:
        add_time(config["senddb"], username, to)
        send_ooo(server, username, msg, config["copy_to_sentmail"])
        log.info("Sent OOO to %s (username %s)", to, username)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: AGPL-3.0-only
from contextlib import closing
import fcntl
import os.path
import shutil
import time

import xapian

"""

index and query mapi fields via python-xapian

useful reading to understand the code:

http://xapian.org/docs/omega/termprefixes.html
http://xapian.org/docs/apidoc/html/classXapian_1_1TermGenerator.html

things would be a lot simpler if we stored everything in 1 xapian database,
but in the old situation we have one database per store which is nice.


"""

class Plugin:
    def __init__(self, index_path, log):
        self.index_path = index_path
        self.log = log
        self.data = []
        self.deletes = []

    def open_db(self, server_guid, store_guid, writable=False, log=None):
        """ open xapian database; if locked, wait until unlocked """

        dbpath = os.path.join(self.index_path, '%s-%s' % (server_guid, store_guid))
        try:
            if writable:
                with open(os.path.join(dbpath+'.lock'), 'w') as lockfile: # avoid compaction using an external lock to be safe
                    fcntl.flock(lockfile.fileno(), fcntl.LOCK_EX)
                    while True:
                        try:
                            return xapian.WritableDatabase(dbpath, xapian.DB_CREATE_OR_OPEN)
                        except xapian.DatabaseLockError:
                            time.sleep(0.1)
            else:
                return xapian.Database(dbpath)
        except xapian.DatabaseOpeningError:
            if log:
                log.warn('could not open database: %s', dbpath)

    def extract_terms(self, text):
        """ extract terms as if we are indexing """
        doc = xapian.Document()
        tg = xapian.TermGenerator()
        tg.set_document(doc)
        text = text.replace('_', ' ') # xapian sees '_' as a word-character (to search for identifiers in source code)
        tg.index_text(text)
        return [t.term.decode('utf-8') for t in doc.termlist()]

    def search(self, server_guid, store_guid, folder_ids, fields_terms, query, limit_results, log):
        """ handle query; see links in the top for a description of the Xapian API """

        db = self.open_db(server_guid, store_guid, log=log)
        if not db:
            return []

        qp = xapian.QueryParser()
        qp.add_prefix("sourcekey", "XK:")
        qp.add_prefix("folderid", "XF:")
        for fields, terms in fields_terms:
            for field in fields:
                qp.add_prefix('mapi%d' % field, "XM%d:" % field)
        log.info('performing query: %s', query)
        qp.set_database(db)
        query = qp.parse_query(query, xapian.QueryParser.FLAG_BOOLEAN|xapian.QueryParser.FLAG_PHRASE|xapian.QueryParser.FLAG_WILDCARD)
        enquire = xapian.Enquire(db)
        enquire.set_query(query)
        matches = []
        for match in enquire.get_mset(0, limit_results or db.get_doccount()): # XXX catch exception if database is being updated?
            matches.append(match.document.get_value(0).decode('ascii'))
        db.close()
        return matches

    def suggest(self, server_guid, store_guid, terms, orig, log):
        """ update original search text with suggested terms """

        db = self.open_db(server_guid, store_guid, log=log)
        if not db:
            return orig

        with closing(db) as db:
            # XXX revisit later. looks like xapian cannot do this for us? :S
            for term in sorted(terms, key=lambda s: len(s), reverse=True):
                suggestion = db.get_spelling_suggestion(term).decode('utf8') or term
                orig = orig.replace(term, suggestion, 1)
            return orig

    def update(self, doc):
        """ new/changed document """

        self.data.append(doc)

    def delete(self, doc):
        """ deleted document """

        self.deletes.append(doc)

    def commit(self, suggestions):
        """ index pending documents; see links in the top for a description of the Xapian API """

        if not self.data and not self.deletes:
            return
        t0 = time.time()
        nitems = len(self.data)
        try:
            # XXX we assume here that all data is from the same store
            doc = (self.data or self.deletes)[0]
            with closing(self.open_db(doc['serverid'], doc['storeid'], writable=True, log=self.log)) as db:
                termgenerator = xapian.TermGenerator()
                termgenerator.set_database(db)
                flags = 0
                if suggestions:
                    flags |= termgenerator.FLAG_SPELLING
                termgenerator.set_flags(flags)
                for doc in self.data:
                    xdoc = xapian.Document()
                    termgenerator.set_document(xdoc)
                    for key, value in doc.items():
                        if key.startswith('mapi'):
                            value = value.replace('_', ' ') # xapian sees '_' as a word-character (to search for identifiers in source code)
                            termgenerator.index_text_without_positions(value) # add to full-text, needed for spelling dict?
                            termgenerator.index_text_without_positions(value, 1, 'XM%s:' % key[4:])
                    xdoc.add_value(0, str(doc['docid']))
                    sourcekey_term = 'XK:'+doc['sourcekey'].lower()
                    xdoc.add_term(sourcekey_term)
                    xdoc.add_term('XF:'+str(doc['folderid'])) #XXX
                    xdoc.set_data(doc['data'])
                    db.replace_document(sourcekey_term, xdoc)
                for doc in self.deletes:
                    db.delete_document('XK:'+doc['sourcekey'].lower())
            self.log.debug('commit took %.2f seconds (%d items)', time.time()-t0, nitems)
        finally:
            self.data = []
            self.deletes = []

    def reindex(self, server_guid, store_guid):
        """ remove database so we can cleanly reindex the store """

        dbpath = os.path.join(self.index_path, '%s-%s' % (server_guid, store_guid))
        self.log.info('removing %s', dbpath)
        shutil.rmtree(dbpath, ignore_errors=True) # may not exist yet (no items to index)



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

from MAPI import (
    ROW_MODIFY,
)

from MAPI.Tags import (
    PR_RULE_NAME_W, PR_RULE_STATE, PR_RULE_PROVIDER_W, PR_RULE_ACTIONS,
    ST_ENABLED, PR_RULE_CONDITION
)

from MAPI.Struct import (
    ACTIONS, ACTION, actMoveCopy, ROWENTRY, SPropValue,
)

from .compat import (
    bdec as _bdec
)

from .restriction import Restriction

class Rule(object):
    """Rule class.

    A rule consist of a :class:`condition <Restriction>` and a set of
    :class:`actions <Action>`. When a mail is delivered (usually in the
    inbox), and it matches the condition, the actions are executed.

    Typical actions are deletion, forwarding, copying and moving.
    """

    def __init__(self, mapirow, table):
        self.mapirow = mapirow
        self.table = table
        #: Rule name.
        self.name = mapirow.get(PR_RULE_NAME_W)
        #: Rule provider.
        self.provider = mapirow.get(PR_RULE_PROVIDER_W)
        if PR_RULE_STATE in mapirow:
            #: Is the rule active.
            self.active = bool(mapirow[PR_RULE_STATE] & ST_ENABLED)
        else:
            self.active = False

    @property
    def restriction(self):
        """Rule :class:`condition <Restriction>`."""
        if PR_RULE_CONDITION in self.mapirow:
            return Restriction(mapiobj=self.mapirow[PR_RULE_CONDITION])

    def create_action(self, type_, folder=None):
        if type_ == 'move': # TODO other types
            action = ACTION(
                1, 0, None, None, 0x0,
                actMoveCopy(_bdec(folder.store.entryid), _bdec(folder.entryid))
            )

            row = []
            found = False
            for proptag, value in self.mapirow.items():
                if proptag == PR_RULE_ACTIONS:
                    found = True
                    value.lpAction.append(action)
                row.append(SPropValue(proptag, value))
            if not found:
                row.append(SPropValue(PR_RULE_ACTIONS, ACTIONS(1, [action])))

            self.table.ModifyTable(0, [ROWENTRY(ROW_MODIFY, row)])

        return Action(action)

    def actions(self):
        """Rule :class:`actions <Action>`."""
        if PR_RULE_ACTIONS in self.mapirow:
            for action in self.mapirow[PR_RULE_ACTIONS].lpAction:
                yield Action(action)

    def __unicode__(self):
        return "Rule('%s')" % self.name

    def __repr__(self):
        return self.__unicode__()

class Action(object):
    """:class:`Rule <Rule>` action class."""

    acttype_str = {
        1: 'move',
        2: 'copy',
        3: 'reply',
        4: 'oof_reply',
        5: 'defer',
        6: 'bounce',
        7: 'forward',
        8: 'delegate',
        9: 'tag',
        10: 'delete',
        11: 'mark_read',
    }

    def __init__(self, mapiobj):
        self.mapiobj = mapiobj
        #: Action operator (*move*, *delete*, ..)
        self.operator = self.acttype_str[mapiobj.acttype]

    def __unicode__(self):
        return "Action('%s')" % self.operator

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-only
from sys import hexversion

from MAPI.Util import GetPublicStore
from MAPI.Struct import NEWMAIL_NOTIFICATION
from MAPI import MAPI_UNICODE, MAPI_MODIFY, OPEN_IF_EXISTS, MDB_WRITE
from MAPI.Tags import (PR_RECEIVED_BY_EMAIL_ADDRESS_W, PR_EC_COMPANY_NAME_W,
                       PR_IPM_PUBLIC_FOLDERS_ENTRYID, PR_ENTRYID,
                       PR_MAILBOX_OWNER_ENTRYID, IID_IMessage, IID_IExchangeManageStore)

from plugintemplates import IMapiDAgentPlugin, MP_CONTINUE, MP_STOP_SUCCESS

import zconfig


class MoveToPublic(IMapiDAgentPlugin):

    prioPreDelivery = 50

    configfile = '/etc/kopano/movetopublic.cfg'

    def __init__(self, logger):
        self.rulelist = {}
        IMapiDAgentPlugin.__init__(self, logger)

        self.Init()

    def Init(self):
        config = zconfig.ZConfigParser(self.configfile, defaultoptions={})

        # scan max for 100 settings
        for i in range(1, 100, 1):
            try:
                data = config.getdict('rule'+str(i), ['recipient', 'destination_folder'])
                self.rulelist[data['recipient'].lower()] = data['destination_folder']
            except:
                break

        self.logger.logDebug("*--- Rule list %s" % self.rulelist)

    def PreDelivery(self, session, addrbook, store, folder, message):

        props = message.GetProps([PR_RECEIVED_BY_EMAIL_ADDRESS_W], 0)
        if props[0].ulPropTag != PR_RECEIVED_BY_EMAIL_ADDRESS_W:
            self.logger.logError("!--- No received by emailaddress")
            return MP_CONTINUE,

        recipient = props[0].Value.lower()

        if recipient not in self.rulelist:
            self.logger.logInfo("*--- No rule for recipient '%s'" % recipient.encode('utf-8'))
            return MP_CONTINUE,

        publicstore = GetPublicStore(session)
        if publicstore is None:
            # check for company public
            companyname = None

            storeprops = store.GetProps([PR_MAILBOX_OWNER_ENTRYID], 0)
            if storeprops[0].ulPropTag == PR_MAILBOX_OWNER_ENTRYID:
                user = addrbook.OpenEntry(storeprops[0].Value, None, 0)
                userprops = user.GetProps([PR_EC_COMPANY_NAME_W], 0)
                if userprops[0].ulPropTag == PR_EC_COMPANY_NAME_W:
                    companyname = userprops[0].Value

                if companyname is None:
                    self.logger.logError("!--- Can not open a public store")
                    return MP_CONTINUE,

                ema = store.QueryInterface(IID_IExchangeManageStore)
                publicstoreid = ema.CreateStoreEntryID(None, companyname, MAPI_UNICODE)
                publicstore = session.OpenMsgStore(0, publicstoreid, None, MDB_WRITE)

        publicfolders = publicstore.OpenEntry(publicstore.GetProps([PR_IPM_PUBLIC_FOLDERS_ENTRYID], 0)[0].Value, None, MAPI_MODIFY)

        folderlist = self.rulelist[recipient].split('/')
        if len(folderlist) == 0:
            self.logger.logWarn("!--- No folders in the rule of recipient '%s'" % recipient.encode('utf-8'))
            return MP_CONTINUE,

        folder = publicfolders
        for foldername in folderlist:
            if len(foldername) > 0:
                if hexversion >= 0x03000000:
                    folder = folder.CreateFolder(0, foldername, "Create by Move to Public plugin", None, OPEN_IF_EXISTS | MAPI_UNICODE)
                else:
                    folder = folder.CreateFolder(0, foldername, "Create by Move to Public plugin", None, OPEN_IF_EXISTS)

        msgnew = folder.CreateMessage(None, 0)
        tags = message.GetPropList(MAPI_UNICODE)
        message.CopyProps(tags, 0, None, IID_IMessage, msgnew, 0)

        msgnew.SaveChanges(0)
        folderid = folder.GetProps([PR_ENTRYID], 0)[0].Value
        msgid = msgnew.GetProps([PR_ENTRYID], 0)[0].Value

        publicstore.NotifyNewMail(NEWMAIL_NOTIFICATION(msgid, folderid, 0, None, 0))

        self.logger.logInfo("*--- Message moved to public folder '%s'" % (self.rulelist[recipient].encode('utf-8')))

        return MP_STOP_SUCCESS,



# SPDX-License-Identifier: AGPL-3.0-only

from wraplogger import WrapLogger
from pluginmanager import PluginManager
from plugintemplates import IMapiDAgentPlugin, IMapiSpoolerPlugin


class DAgentPluginManager(object):

    def __init__(self, plugindir):
        self.logger = WrapLogger()
        self.pluginmanager = PluginManager(plugindir, self.logger)

        self.pluginmanager.loadPlugins(IMapiDAgentPlugin)

    def PostConverting(self, session, addrbook, store, folder, message):
        return self.pluginmanager.processPluginFunction('PostConverting', session, addrbook, store, folder, message)

    def PreDelivery(self, session, addrbook, store, folder, message):
        return self.pluginmanager.processPluginFunction('PreDelivery', session, addrbook, store, folder, message)

    def PostDelivery(self, session, addrbook, store, folder, message):
        return self.pluginmanager.processPluginFunction('PostDelivery', session, addrbook, store, folder, message)

    def PreRuleProcess(self, session, addrbook, store, rulestable):
        return self.pluginmanager.processPluginFunction('PreRuleProcess', session, addrbook, store, rulestable)

    def SendNewMailNotify(self, session, addrbook, store, folder, message):
        return self.pluginmanager.processPluginFunction('SendNewMailNotify', session, addrbook, store, folder, message)


class SpoolerPluginManager(object):

    def __init__(self, plugindir):
        self.logger = WrapLogger()
        self.pluginmanager = PluginManager(plugindir, self.logger)

        self.pluginmanager.loadPlugins(IMapiSpoolerPlugin)

    def PreSending(self, session, addrbook, store, folder, message):
        return self.pluginmanager.processPluginFunction('PreSending', session, addrbook, store, folder, message)



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI.Tags import (
    PR_ATTACHMENT_CONTACTPHOTO, PR_GIVEN_NAME_W, PR_MIDDLE_NAME_W,
    PR_SURNAME_W, PR_NICKNAME_W, PR_TITLE_W, PR_GENERATION_W,
    PR_COMPANY_NAME_W, PR_MOBILE_TELEPHONE_NUMBER_W, PR_CHILDRENS_NAMES_W,
    PR_BIRTHDAY, PR_SPOUSE_NAME_W, PR_INITIALS_W, PR_DISPLAY_NAME_PREFIX_W,
    PR_DEPARTMENT_NAME_W, PR_OFFICE_LOCATION_W, PR_PROFESSION_W,
    PR_MANAGER_NAME_W, PR_ASSISTANT_W, PR_BUSINESS_HOME_PAGE_W,
    PR_HOME_TELEPHONE_NUMBER_W, PR_HOME2_TELEPHONE_NUMBER_W,
    PR_BUSINESS_TELEPHONE_NUMBER_W, PR_BUSINESS2_TELEPHONE_NUMBER_W,
    PR_HOME_ADDRESS_STREET_W, PR_HOME_ADDRESS_CITY_W,
    PR_HOME_ADDRESS_POSTAL_CODE_W, PR_HOME_ADDRESS_STATE_OR_PROVINCE_W,
    PR_HOME_ADDRESS_COUNTRY_W, PR_OTHER_ADDRESS_STREET_W,
    PR_OTHER_ADDRESS_CITY_W, PR_OTHER_ADDRESS_POSTAL_CODE_W,
    PR_OTHER_ADDRESS_STATE_OR_PROVINCE_W, PR_OTHER_ADDRESS_COUNTRY_W,
)

from .pidlid import (
    PidLidEmail1AddressType, PidLidEmail1DisplayName, PidLidEmail1EmailAddress,
    PidLidEmail1OriginalEntryId, PidLidEmail2AddressType,
    PidLidEmail2DisplayName,
    PidLidEmail2EmailAddress, PidLidEmail2OriginalEntryId, PidLidYomiFirstName,
    PidLidEmail3AddressType, PidLidEmail3DisplayName, PidLidEmail3EmailAddress,
    PidLidEmail3OriginalEntryId, PidLidYomiLastName, PidLidYomiCompanyName,
    PidLidFileUnder, PidLidInstantMessagingAddress, PidLidWorkAddressStreet,
    PidLidWorkAddressCity, PidLidWorkAddressState, PidLidWorkAddressPostalCode,
    PidLidWorkAddressCountry,
)


from .address import Address
from .errors import NotFoundError
from .picture import Picture

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

class PhysicalAddress(object):
    """PhysicalAddress class

    Abstraction for physical addresses.
    """
    def __init__(self, item, proptags):
        street, city, postal_code, state, country = proptags
        #: Street
        self.street = item.get(street, '')
        #: City
        self.city = item.get(city, '')
        #: Postal code
        self.postal_code = item.get(postal_code, '')
        #: State
        self.state = item.get(state, '')
        #: Country
        self.country = item.get(country, '')

    def __repr__(self):
        return self.__unicode__()

    def __unicode__(self):
        return 'PhysicalAddress()'

class Contact(object):
    """Contact mixin class

    Contact-specific functionality, mixed into the :class:`Item`
    class.
    """

    @property
    def email(self):
        """Primary email address."""
        return self.email1

    @email.setter
    def email(self, addr):
        self.email1 = addr

    @property
    def email1(self):
        if self.address1:
            return self.address1.email

    # TODO name1 (setter) etc..?

    @email1.setter
    def email1(self, addr): # TODO should be email address?
        self.address1 = addr

    @property
    def address1(self):
        """Primary :class:`address <Address>`."""
        return Address(
            self.server,
            self.get(PidLidEmail1AddressType),
            self.get(PidLidEmail1DisplayName),
            self.get(PidLidEmail1EmailAddress),
            self.get(PidLidEmail1OriginalEntryId),
        )

    @address1.setter
    def address1(self, addr):
        pr_addrtype, pr_dispname, pr_email, pr_entryid = \
            self._addr_props(addr)

        self[PidLidEmail1AddressType] = str(pr_addrtype)
        self[PidLidEmail1DisplayName] = pr_dispname
        self[PidLidEmail1EmailAddress] = pr_email
        self[PidLidEmail1OriginalEntryId] = pr_entryid

    def addresses(self): # TODO multiple
        """Return all :class:`addresses <Address>`."""
        yield self.address1

    @property
    def email2(self):
        """Secondary email address."""
        if self.address1:
            return self.address2.email

    @email2.setter
    def email2(self, addr):  # TODO should be email address?
        self.address2 = addr

    @property
    def address2(self):
        """Secondary :class:`address <Address>`."""
        return Address(
            self.server,
            self.get(PidLidEmail2AddressType),
            self.get(PidLidEmail2DisplayName),
            self.get(PidLidEmail2EmailAddress),
            self.get(PidLidEmail2OriginalEntryId),
        )

    @address2.setter
    def address2(self, addr):
        pr_addrtype, pr_dispname, pr_email, pr_entryid = \
            self._addr_props(addr)

        self[PidLidEmail2AddressType] = str(pr_addrtype)
        self[PidLidEmail2DisplayName] = pr_dispname
        self[PidLidEmail2EmailAddress] = pr_email
        self[PidLidEmail2OriginalEntryId] = pr_entryid

    @property
    def email3(self):
        """Tertiary email address."""
        if self.address3:
            return self.address3.email

    @email3.setter
    def email3(self, addr):  # TODO should be email address?
        self.address3 = addr

    @property
    def address3(self):
        """Tertiary :class:`address <Address>`."""
        return Address(
            self.server,
            self.get(PidLidEmail3AddressType),
            self.get(PidLidEmail3DisplayName),
            self.get(PidLidEmail3EmailAddress),
            self.get(PidLidEmail3OriginalEntryId),
        )

    @address3.setter
    def address3(self, addr):
        pr_addrtype, pr_dispname, pr_email, pr_entryid = \
            self._addr_props(addr)

        self[PidLidEmail3AddressType] = str(pr_addrtype)
        self[PidLidEmail3DisplayName] = pr_dispname
        self[PidLidEmail3EmailAddress] = pr_email
        self[PidLidEmail3OriginalEntryId] = pr_entryid

    # TODO uniformize with user.photo? class Picture?
    @property
    def photo(self):
        """Contact :class:`photo <Picture>`"""
        for attachment in self.attachments():
            if attachment.get(PR_ATTACHMENT_CONTACTPHOTO):
                return Picture(data=attachment.data, name=attachment.name,
                    mimetype=attachment.mimetype)

    def set_photo(self, name, data, mimetype):
        """Set contact :class:`photo <Picture>`

        :param name: file name
        :param data: file binary data
        :param mimetype: file MIME type
        """

        for attachment in self.attachments():
            if attachment.get(PR_ATTACHMENT_CONTACTPHOTO):
                self.delete(attachment)
        attachment = self.create_attachment(name, data)
        attachment[PR_ATTACHMENT_CONTACTPHOTO] = True
        attachment.mimetype = mimetype
        _utils._save(self.mapiobj)
        return attachment

    @property
    def initials(self):
        """Initials."""
        return self.get(PR_INITIALS_W, '')

    @property
    def first_name(self):
        """First name."""
        return self.get(PR_GIVEN_NAME_W, '')

    @property
    def middle_name(self):
        """Middle name."""
        return self.get(PR_MIDDLE_NAME_W, '')

    @property
    def last_name(self):
        """Last name."""
        return self.get(PR_SURNAME_W, '')

    @property
    def nickname(self):
        """Nickname."""
        return self.get(PR_NICKNAME_W, '')

    @property
    def title(self):
        """Title."""
        return self.get(PR_DISPLAY_NAME_PREFIX_W, '')

    @property
    def generation(self):
        """Generation."""
        return self.get(PR_GENERATION_W, '')

    @property
    def company_name(self):
        """Company name."""
        return self.get(PR_COMPANY_NAME_W, '')

    @property
    def children(self):
        """List of children names."""
        try:
            return self[PR_CHILDRENS_NAMES_W]
        except NotFoundError:
            return []

    @property
    def spouse(self):
        """Spouse."""
        return self.get(PR_SPOUSE_NAME_W, '')

    @property
    def birthday(self):
        """Birthday."""
        return self.get(PR_BIRTHDAY)

    @property
    def yomi_first_name(self):
        """Yomi (phonetic) first name."""
        return self.get(PidLidYomiFirstName, '')

    @property
    def yomi_last_name(self):
        """Yomi (phonetic) last name."""
        return self.get(PidLidYomiLastName, '')

    @property
    def yomi_company_name(self):
        """Yomi (phonetic) company name."""
        return self.get(PidLidYomiCompanyName, '')

    @property
    def file_as(self):
        """File as."""
        return self.get(PidLidFileUnder, '')

    @property
    def job_title(self):
        """Job title."""
        return self.get(PR_TITLE_W, '')

    @property
    def department(self):
        """Department."""
        return self.get(PR_DEPARTMENT_NAME_W, '')

    @property
    def office_location(self):
        """Office location."""
        return self.get(PR_OFFICE_LOCATION_W, '')

    @property
    def profession(self):
        """Profession."""
        return self.get(PR_PROFESSION_W, '')

    @property
    def manager(self):
        """Manager."""
        return self.get(PR_MANAGER_NAME_W, '')

    @property
    def assistant(self):
        """Assistant."""
        return self.get(PR_ASSISTANT_W, '')

    @property
    def business_homepage(self):
        """Business homepage."""
        return self.get(PR_BUSINESS_HOME_PAGE_W, '')

    @property
    def mobile_phone(self):
        """Mobile phone."""
        return self.get(PR_MOBILE_TELEPHONE_NUMBER_W, '')

    def _str_list(self, proptags):
        result = []
        for proptag in proptags:
            val = self.get(proptag)
            if val:
                result.append(val)
        return result

    @property
    def home_phones(self):
        """List of home phone numbers."""
        return self._str_list(
            [PR_HOME_TELEPHONE_NUMBER_W, PR_HOME2_TELEPHONE_NUMBER_W])

    @property
    def business_phones(self):
        """List of business phone numbers."""
        return self._str_list(
            [PR_BUSINESS_TELEPHONE_NUMBER_W, PR_BUSINESS2_TELEPHONE_NUMBER_W])

    @property
    def im_addresses(self):
        """List of instant messaging addresses."""
        return self._str_list([PidLidInstantMessagingAddress])

    @property
    def home_address(self):
        """Home :class:`address <PhysicalAddress>`."""
        return PhysicalAddress(self, [
            PR_HOME_ADDRESS_STREET_W,
            PR_HOME_ADDRESS_CITY_W,
            PR_HOME_ADDRESS_POSTAL_CODE_W,
            PR_HOME_ADDRESS_STATE_OR_PROVINCE_W,
            PR_HOME_ADDRESS_COUNTRY_W,
        ])

    @property
    def business_address(self):
        """Business :class:`address <PhysicalAddress>`."""
        # TODO why not PR_BUSINESS_ADDRESS_*..?
        return PhysicalAddress(self, [
            PidLidWorkAddressStreet,
            PidLidWorkAddressCity,
            PidLidWorkAddressPostalCode,
            PidLidWorkAddressState,
            PidLidWorkAddressCountry,
        ])

    @property
    def other_address(self):
        """Other :class:`address <PhysicalAddress>`."""
        return PhysicalAddress(self, [
            PR_OTHER_ADDRESS_STREET_W,
            PR_OTHER_ADDRESS_CITY_W,
            PR_OTHER_ADDRESS_POSTAL_CODE_W,
            PR_OTHER_ADDRESS_STATE_OR_PROVINCE_W,
            PR_OTHER_ADDRESS_COUNTRY_W,
        ])



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2017 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import (
    PT_BINARY, PT_UNICODE, PT_ERROR, MAPI_E_NOT_FOUND
)

from MAPI.Defs import PROP_TYPE
from MAPI.Struct import SPropValue

from .errors import NotFoundError

from . import property_ as _prop

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

class Properties(object):
    """Property mixin class

    MAPI property-specific functionality, mixed into several
    classes whose instances represent sets of MAPI properties,
    such as :class:`Item <Item>` and :class:`Folder <Folder>`.
    """

    def prop(self, proptag, create=False, proptype=None):
        """Return :class:`property <Property>` with given property tag.

        :param proptag: MAPI property tag
        :param create: create property if it doesn't exist
        """
        return _prop.prop(self, self.mapiobj, proptag, create=create,
                          proptype=proptype)

    def get_prop(self, proptag):
        """Return :class:`property <Property>` with given proptag or
        *None* if not found.

        :param proptag: MAPI property tag
        """
        try:
            return self.prop(proptag)
        except NotFoundError:
            pass

    def create_prop(self, proptag, value, proptype=None):
        """Create :class:`property <Property>` with given proptag.

        :param proptag: MAPI property tag
        :param value: property value (or a default value is used)
        """
        return _prop.create_prop(self, self.mapiobj, proptag, value, proptype)

    def props(self, namespace=None):
        """Return all :class:`properties <Property>`."""
        return _prop.props(self.mapiobj, namespace)

    # mapi objects/properties are basically key-value stores
    # so the following provides some useful dict-like behaviour

    def get(self, proptag, default=None):
        """Return :class:`property <Property>` value for given proptag or
        *None* if property does not exist.

        :param proptag: MAPI property tag
        """
        try:
            return self.prop(proptag).value
        except NotFoundError:
            return default

    def __getitem__(self, proptag):
        """Return :class:`property <Property>` value for given proptag."""
        return self.prop(proptag).value

    def __setitem__(self, proptag, value):
        """Set :class:`property <Property>` value for given proptag,
        creating the property if it doesn't exist.
        """
        self.prop(proptag, create=True).value = value

    def __delitem__(self, proptag):
        """Delete the :class:`property <Property>` with given proptag."""
        self.delete(self.prop(proptag))

    # the following is faster in case of preloaded/cached table data,
    # because it avoids the backend preloading all item properties with
    # about 20 SQL statements _per item_ (!)

    # TODO generalize for any property?
    def _get_fast(self, proptag, default=None, must_exist=False, capped=False):
        # in cache

        prop = self._cache.get(proptag)
        if prop is not None:
            proptype = PROP_TYPE(prop.proptag)
            value = prop.value

            if proptype == PT_ERROR and value == MAPI_E_NOT_FOUND:
                return default

            # mapi table cells are limited to 255 characters/bytes
            # TODO check other types
            if capped or not (proptype in (PT_UNICODE, PT_BINARY) and len(value) >= 255):
                return value

        # fallback to (slow) lookup
        try:
            return self.prop(proptag).value
        except NotFoundError:
            if must_exist:
                raise
            else:
                return default

    def _set_fast(self, proptag, value):
        self._cache.pop(proptag, None)
        self.mapiobj.SetProps([SPropValue(proptag, value)])
        _utils._save(self.mapiobj)

    def __repr__(self):
        return self.__unicode__()



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-or-later
from .version import __version__

from contextlib import closing
import grp
import os
import sys
import time

import bsddb3 as bsddb

import kopano
from kopano import Config, log_exc

"""
kopano-spamd - ICS driven spam learning daemon for Kopano / SpamAssasin
"""

CONFIG = {
    'spam_dir': Config.string(default="/var/db/kopano/spamd/spam"),
    'ham_dir': Config.string(default="/var/db/kopano/spamd/ham"),
    'spam_db': Config.string(default="/var/db/kopano/spamd/spam.db"),
    'sa_group': Config.string(default="_spamdaemon"),
    'run_as_user': Config.string(default="_kopano"),
    'run_as_group': Config.string(default="_kopano"),
    'learn_ham': Config.boolean(default=True),
    'header_tag': Config.string(default="x-spam-flag")
}


class Service(kopano.Service):
    def main(self):
        server = self.server
        state = server.state # start from current state
        importer = Importer(self)
        with log_exc(self.log):
            while True:
                state = server.sync(importer, state)
                time.sleep(1)


class Importer:
    def __init__(self, service):
        self.log = service.log
        self.spamdb = service.config['spam_db']
        self.spamdir = service.config['spam_dir']
        self.hamdir = service.config['ham_dir']
        self.sagroup = service.config['sa_group']
        self.learnham = service.config['learn_ham']
        self.headertag = service.config['header_tag'].lower()

    def mark_spam(self, searchkey):
        if not isinstance(searchkey, bytes): # python3
            searchkey = searchkey.encode('ascii')
        with closing(bsddb.btopen(self.spamdb, 'c')) as db:
            db[searchkey] = ''

    def was_spam(self, searchkey):
        if not isinstance(searchkey, bytes): # python3
            searchkey = searchkey.encode('ascii')
        with closing(bsddb.btopen(self.spamdb, 'c')) as db:
            return searchkey in db

    def update(self, item, flags):
        with log_exc(self.log):
            if item.message_class != 'IPM.Note': # TODO None?
                return

            searchkey = item.searchkey
            header = item.header(self.headertag)

            if (item.folder == item.store.junk and \
                (not header or header.upper() != 'YES')):

                fn = os.path.join(self.hamdir, searchkey + '.eml')
                if os.path.isfile(fn):
                    os.unlink(fn)

                self.log.info("Learning message as SPAM, entryid: %s", item.entryid)
                self.learn(item, searchkey, True)

            elif (item.folder == item.store.inbox and \
                  self.learnham and \
                  self.was_spam(searchkey)):

                fn = os.path.join(self.spamdir, searchkey + '.eml')
                if os.path.isfile(fn):
                    os.unlink(fn)

                self.log.info("Learning message as HAM, entryid: %s", item.entryid)
                self.learn(item, searchkey, False)

    def learn(self, item, searchkey, spam):
        spameml = item.eml()
        dir_ = self.spamdir if spam else self.hamdir
        emlfilename = os.path.join(dir_, searchkey + '.eml')

        with closing(open(emlfilename, "wb")) as fh:
            fh.write(spameml)

        uid = os.getuid()
        gid = grp.getgrnam(self.sagroup).gr_gid
        os.chown(emlfilename, uid, gid)
        os.chmod(emlfilename, 0o660)

        if spam:
            self.mark_spam(searchkey)

def main():
    parser = kopano.parser('CKQSFl')  # select common cmd-line options
    options = parser.parse_args()[0]
    service = Service('spamd', config=CONFIG, options=options)
    if service.config['learn_ham'] == True and not os.path.exists(service.config['ham_dir']):
        os.makedirs(service.config['ham_dir'])
    service.start()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: AGPL-3.0-only

MP_CONTINUE         = 0
MP_FAILED           = 1
MP_STOP_SUCCESS     = 2
MP_STOP_FAILED      = 3
MP_EXIT             = 4
MP_RETRY_LATER      = 5


class IMapiDAgentPlugin(object):
    prioPostConverting = 9999
    prioPreDelivery = 9999
    prioPostDelivery = 9999
    prioPreRuleProcess = 9999
    prioSendNewMailNotify = 9999

    def __init__(self, logger):
        self.logger = logger

    def PostConverting(self, session, addrbook, store, folder, message):
        raise NotImplementedError('PostConverting not implemented, function ignored')

    def PreDelivery(self, session, addrbook, store, folder, message):
        raise NotImplementedError('PreDelivery not implemented, function ignored')

    def PostDelivery(self, session, addrbook, store, folder, message):
        raise NotImplementedError('PostDelivery not implemented, function ignored')

    def PreRuleProcess(self, session, addrbook, store, rulestable):
         raise NotImplementedError('PreRuleProcess not implemented, function ignored')

    def SendNewMailNotify(self, session, addrbook, store, folder, message):
         raise NotImplementedError('SendNewMailNotify not implemented, function ignored')


class IMapiSpoolerPlugin(object):
    prioPreSending = 9999

    def __init__(self, logger):
        self.logger = logger

    def PreSending(self, session, addrbook, store, folder, message):
        raise NotImplementedError('PreSending not implemented, function ignored')



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import calendar
import datetime
import sys

from MAPI import (
    PT_ERROR, PT_BINARY, PT_MV_BINARY, PT_UNICODE, PT_LONG, PT_OBJECT,
    PT_BOOLEAN, MV_FLAG, PT_SHORT, PT_MV_SHORT, PT_LONGLONG,
    PT_MV_LONG, PT_FLOAT, PT_MV_FLOAT, PT_DOUBLE, PT_MV_DOUBLE, PT_STRING8,
    PT_CURRENCY, PT_MV_CURRENCY, PT_APPTIME, PT_MV_APPTIME, PT_MV_LONGLONG,
    PT_SYSTIME, MAPI_E_NOT_ENOUGH_MEMORY, PT_MV_STRING8,
    PT_MV_UNICODE, PT_MV_SYSTIME, PT_CLSID, PT_MV_CLSID, MNID_STRING,
    MAPI_E_NOT_FOUND, MNID_ID, MAPI_UNICODE,
)
from MAPI.Defs import (
    PROP_ID, PROP_TYPE, HrGetOneProp,
    CHANGE_PROP_TYPE,
)
from MAPI.Struct import (
    SPropValue, MAPIErrorNotFound, MAPIErrorNoSupport,
    MAPIErrorNotEnoughMemory
)
import MAPI.Tags
from MAPI.Tags import (
    PR_BODY_W, PR_HTML, PR_RTF_COMPRESSED, PR_RTF_IN_SYNC, PR_NULL
)
from MAPI.Time import unixtime

from .defs import (
    REV_TAG, REV_TYPE, GUID_NAMESPACE, MAPINAMEID, NAMESPACE_GUID, STR_GUID,
)
from .compat import (
    benc as _benc,
)
from .errors import Error, NotFoundError

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']
from . import timezone as _timezone

TYPEMAP = {
    'PT_SHORT': PT_SHORT,
    'PT_MV_SHORT': PT_MV_SHORT,
    'PT_LONG': PT_LONG,
    'PT_MV_LONG': PT_MV_LONG,
    'PT_FLOAT': PT_FLOAT,
    'PT_MV_FLOAT': PT_MV_FLOAT,
    'PT_DOUBLE': PT_DOUBLE,
    'PT_MV_DOUBLE': PT_MV_DOUBLE,
    'PT_CURRENCY': PT_CURRENCY,
    'PT_MV_CURRENCY': PT_MV_CURRENCY,
    'PT_APPTIME': PT_APPTIME,
    'PT_MV_APPTIME': PT_MV_APPTIME,
    'PT_BOOLEAN': PT_BOOLEAN,
    'PT_OBJECT': PT_OBJECT,
    'PT_LONGLONG': PT_LONGLONG,
    'PT_MV_LONGLONG': PT_MV_LONGLONG,
    'PT_STRING8': PT_STRING8,
    'PT_MV_STRING8': PT_MV_STRING8,
    'PT_UNICODE': PT_UNICODE,
    'PT_MV_UNICODE': PT_MV_UNICODE,
    'PT_SYSTIME': PT_SYSTIME,
    'PT_MV_SYSTIME': PT_MV_SYSTIME,
    'PT_CLSID': PT_CLSID,
    'PT_MV_CLSID': PT_MV_CLSID,
    'PT_BINARY': PT_BINARY,
    'PT_MV_BINARY': PT_MV_BINARY,
}

 # TODO we may want to use the swigged version in libcommon, once available?
def bestbody(mapiobj):
    # apparently standardized method for determining original message type!
    tag = PR_NULL
    props = mapiobj.GetProps(
        [PR_BODY_W, PR_HTML, PR_RTF_COMPRESSED, PR_RTF_IN_SYNC], 0)

    if (props[3].ulPropTag != PR_RTF_IN_SYNC): # TODO why..
        return tag

    # MAPI_E_NOT_ENOUGH_MEMORY indicates the property exists,
    # but has to be streamed
    if((props[0].ulPropTag == PR_BODY_W or \
        (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and \
         props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and \
       (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and \
        props[1].Value == MAPI_E_NOT_FOUND) and \
       (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and \
        props[2].Value == MAPI_E_NOT_FOUND)):
        tag = PR_BODY_W

    # TODO why not just check MAPI_E_NOT_FOUND..?
    elif((props[1].ulPropTag == PR_HTML or \
          (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and \
           props[1].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and \
         (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and \
          props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY) and \
         (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and \
          props[2].Value == MAPI_E_NOT_ENOUGH_MEMORY) and \
         not props[3].Value):
        tag = PR_HTML

    elif((props[2].ulPropTag == PR_RTF_COMPRESSED or \
          (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and \
           props[2].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and
         (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and \
          props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY) and
         (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and \
          props[1].Value == MAPI_E_NOT_FOUND) and \
         props[3].Value):
        tag = PR_RTF_COMPRESSED

    return tag

def _split_proptag(proptag): # TODO syntax error exception
    parts = proptag.split(':')
    if len(parts) == 2:
        namespace, name = parts
        return None, namespace, name
    else:
        proptype, namespace, name = parts
        return TYPEMAP[proptype], namespace, name

def _name_to_proptag(proptag, mapiobj, proptype=None):
    ptype, namespace, name = _split_proptag(proptag)
    proptype = proptype or ptype

    if name.isdigit(): # TODO
        name = int(name)
    elif name.startswith('0x'):
        name = int(name, 16)

    if namespace in NAMESPACE_GUID:
        guid = NAMESPACE_GUID[namespace]
    elif namespace in STR_GUID:
        guid = STR_GUID[namespace]

    if proptype is None:
        return None, proptype, namespace, name

    nameid = MAPINAMEID(guid, MNID_ID if isinstance(name, int) else \
        MNID_STRING, name)
    # TODO MAPI_CREATE, or too dangerous because of potential db overflow?
    lpname = mapiobj.GetIDsFromNames([nameid], 0)
    proptag = CHANGE_PROP_TYPE(lpname[0], proptype)

    return proptag, proptype, namespace, name

# TODO self
def create_prop(self, mapiobj, proptag, value=None, proptype=None):
    if isinstance(proptag, int) or \
       (isinstance(proptag, str) and ':' not in proptag):
        if isinstance(proptag, str):
            proptag2 = getattr(MAPI.Tags, proptag)
        else:
            proptag2 = proptag
        proptype2 = proptype or PROP_TYPE(proptag)

    else: # named property
        proptag2, proptype2, _, _ = \
            _name_to_proptag(proptag, mapiobj, proptype)

        if proptype2 is None:
            # TODO exception too general?
            raise Error('Missing type to create named property')

    if value is None:
        if proptype2 in (PT_STRING8, PT_UNICODE):
            value = ''
        elif proptype2 == PT_BINARY:
            value = b''
        elif proptype2 == PT_SYSTIME:
            value = unixtime(0)
        elif proptype2 & MV_FLAG:
            value = []
        else:
            value = 0
    else:
        if proptype2 == PT_SYSTIME:
            if value.tzinfo is None:
                value = _timezone._to_utc(value, _timezone.LOCAL)
            else:
                value = value.astimezone(_timezone.UTC)
            value = unixtime(calendar.timegm(value.utctimetuple()))

    # handle invalid type versus value.
    # For example proptype=PT_UNICODE and value=True
    try:
        mapiobj.SetProps([SPropValue(proptag2, value)])
        _utils._save(mapiobj)
    except TypeError:
        raise Error(
            'Could not create property, type and value did not match')

    return prop(self, mapiobj, proptag, proptype=proptype2)

# TODO self
def prop(self, mapiobj, proptag, create=False, value=None, proptype=None):
    if isinstance(proptag, int) or \
       (isinstance(proptag, str) and ':' not in proptag):
        # search for property
        if isinstance(proptag, str):
            proptag = getattr(MAPI.Tags, proptag)
        try:
            sprop = HrGetOneProp(mapiobj, proptag)
        except MAPIErrorNotEnoughMemory:
            data = _utils.stream(mapiobj, proptag)
            sprop = SPropValue(proptag, data)
        except MAPIErrorNotFound:
            # not found, create it?
            if create:
                return create_prop(
                    self, mapiobj, proptag, value=value, proptype=proptype)
            else:
                raise NotFoundError('no such property: %s' %
                    REV_TAG.get(proptag, hex(proptag)))
        return Property(mapiobj, sprop)

    else: # named property
        proptag2, proptype2, namespace, name = \
            _name_to_proptag(proptag, mapiobj, proptype)

        # search for property
        if proptype2:
            try:
                # TODO merge two main branches?
                sprop = HrGetOneProp(mapiobj, proptag2)
                return Property(mapiobj, sprop)
            except MAPIErrorNotEnoughMemory:
                data = _utils.stream(mapiobj, proptag2)
                sprop = SPropValue(proptag2, data)
                return Property(mapiobj, sprop)
            except MAPIErrorNotFound:
                pass
        else:
            # TODO sloow, streaming? default pidlid type-db?
            for prop in self.props(namespace=namespace):
                if prop.name == name:
                    return prop

        # not found, create it?
        if create:
            return create_prop(
                self, mapiobj, proptag, value=value, proptype=proptype)
        else:
            raise NotFoundError('no such property: %s' % proptag)

def props(mapiobj, namespace=None):
    proptags = mapiobj.GetPropList(MAPI_UNICODE)
    sprops = mapiobj.GetProps(proptags, MAPI_UNICODE)
    sprops = [s for s in sprops if not \
        (PROP_TYPE(s.ulPropTag) == PT_ERROR and s.Value == MAPI_E_NOT_FOUND)]
    props = [Property(mapiobj, sprop) for sprop in sprops]

    def prop_key(prop): # sort identically across servers
        if prop.named:
            return (prop.guid, prop.kind, prop.name)
        else:
            return ('', prop.proptag)

    for p in sorted(props, key=prop_key):
        if not namespace or p.namespace == namespace:
            yield p

# do not get full property contents, only when explicitly requested
class SPropDelayedValue(SPropValue):
    def __init__(self, mapiobj, proptag):
        self.mapiobj = mapiobj
        self.ulPropTag = proptag
        self._Value = None

    @property
    def Value(self):
        if self._Value is None:
            try:
                self._Value = _utils.stream(self.mapiobj, self.ulPropTag)
            # TODO eg normalized subject streaming broken..?
            except MAPIErrorNotFound:
                self._Value = None
        return self._Value

class Property(object):
    """Property class

    Low-level abstraction for MAPI properties.
    """

    def __init__(self, parent_mapiobj, mapiobj):
        self._parent_mapiobj = parent_mapiobj
        #: MAPI proptag, for example 0x37001f for PR_SUBJECT_W
        self.proptag = mapiobj.ulPropTag

        if (PROP_TYPE(mapiobj.ulPropTag) == PT_ERROR and \
            mapiobj.Value == MAPI_E_NOT_ENOUGH_MEMORY):
            # avoid slow guessing
            if PROP_ID(self.proptag) == PROP_ID(PR_BODY_W):
                self.proptag = PR_BODY_W
                mapiobj = SPropDelayedValue(parent_mapiobj, self.proptag)
            elif PROP_ID(self.proptag) in \
                (PROP_ID(PR_RTF_COMPRESSED), PROP_ID(PR_HTML)):
                self.proptag = CHANGE_PROP_TYPE(self.proptag, PT_BINARY)
                mapiobj = SPropDelayedValue(parent_mapiobj, self.proptag)
            else: # TODO possible to use above trick to infer all proptags?
                # TODO slow, incomplete?
                for proptype in (PT_BINARY, PT_UNICODE):
                    proptag = (mapiobj.ulPropTag & 0xffff0000) | proptype
                    try:
                        # TODO: Unicode issue?? calls GetProps([proptag], 0)
                        HrGetOneProp(parent_mapiobj, proptag)
                        # TODO how did we end up here, why is this possible
                        self.proptag = proptag
                    except MAPIErrorNotEnoughMemory:
                        mapiobj = SPropDelayedValue(parent_mapiobj, proptag)
                        self.proptag = proptag
                        break
                    except MAPIErrorNotFound:
                        pass

        self.mapiobj = mapiobj

        self._value = None
        self.__lpname = None

    @property
    def id_(self):
        """Id part of proptag, for example 0x37 for PR_SUBJECT_W."""
        return PROP_ID(self.proptag)

    @property
    def idname(self):
        """MAPI proptag name, for example *PR_SUBJECT_W*."""
        return REV_TAG.get(self.proptag)

    @property
    def type_(self):
        """Proptag type, for example 0x1f for PR_SUBJECT_W."""
        return PROP_TYPE(self.proptag)

    @property
    def typename(self):
        """Proptag type name, for example *PT_UNICODE* for PR_SUBJECT_W."""
        return REV_TYPE.get(self.type_)

    @property
    def _lpname(self):
        if self.__lpname is None and self.id_ >= 0x8000:
            try:
                self.__lpname = self._parent_mapiobj.GetNamesFromIDs(
                    [self.proptag], None, 0)[0]
            except MAPIErrorNoSupport: # TODO user.props()?
                pass
        return self.__lpname

    @property
    def named(self):
        """Is the property a named property."""
        lpname = self._lpname
        return bool(lpname)

    @property
    def guid(self):
        """For a named property, return the GUID."""
        lpname = self._lpname
        if lpname:
            return _benc(lpname.guid)

    @property
    def namespace(self):
        """For a named property, return the readable namespace."""
        lpname = self._lpname
        if lpname:
            return GUID_NAMESPACE.get(lpname.guid)

    @property
    def name(self):
        """For a named property, return the name."""
        lpname = self._lpname
        if lpname:
            return lpname.id

    @property
    def kind(self):
        """For a named property, return the name kind."""
        lpname = self._lpname
        if lpname:
            return lpname.kind

    @property
    def kindname(self):
        """For a named property, return the readable name kind."""
        lpname = self._lpname
        if lpname:
            return 'MNID_STRING' if lpname.kind == MNID_STRING else 'MNID_ID'

    @property
    def value(self):
        """Property value.

        For PT_SYSTIME properties, convert to/from timezone-unaware
        system-local datetimes.
        """
        if self._value is None:
            if self.type_ == PT_SYSTIME: # TODO generalize, property?
                # TODO add global flag to enable tz-aware UTC datetimes
                try:
                    value = datetime.datetime.utcfromtimestamp(self.mapiobj.Value.unixtime)
                    value = _timezone._from_utc(value, _timezone.LOCAL)
                except ValueError: # Y10K: datetime is limited to 4-digit years
                    value = datetime.datetime(9999, 1, 1)
                self._value = value
            else:
                self._value = self.mapiobj.Value
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if self.type_ == PT_SYSTIME:
            # Ensure that time is stored as UTC.
            if value.tzinfo is None:
                value = _timezone._to_utc(value, _timezone.LOCAL)
            else:
                value = value.astimezone(_timezone.UTC)
            value = unixtime(calendar.timegm(value.utctimetuple()))
        self._parent_mapiobj.SetProps([SPropValue(self.proptag, value)])
        _utils._save(self._parent_mapiobj)

    @property
    def strid(self):
        """Readable identifier for named/unnamed property."""
        if self.named:
            return '%s:%s' % (self.namespace or self.guid, self.name)
        else:
            return self.idname or hex(self.proptag)

    @property
    def strval(self):
        """String representation of property value."""
        def flatten(v):
            if isinstance(v, list):
                return ','.join(flatten(e) for e in v)
            elif isinstance(v, bool):
                return str(v)
            elif isinstance(v, datetime.datetime):
                return str(v.isoformat(' '))
            elif v is None:
                return ''
            elif (self.type_ in (PT_BINARY, PT_MV_BINARY) or \
                  isinstance(v, bytes)):
                return _benc(v)
            else:
                return str(v)
        return flatten(self.value)

    def __unicode__(self):
        return 'Property(%s)' % self.strid

    # TODO: check if data is binary and convert it to hex
    def __repr__(self):
        return self.__unicode__()



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-or-later

import collections
from datetime import datetime, timedelta
import os
import sys
import kopano

from MAPI.Tags import PR_DISPLAY_TYPE_EX, PR_EMS_AB_ROOM_CAPACITY

RECURRENCE_AVAILABILITY_RANGE = 180  # days
DT_EQUIPMENT = 8

# XXX ZCP-9901 still relevant without outlook?


def capacity(user):  # XXX pyko?
    """ equipment resources can be overbooked up to N times """

    capacity = os.getenv('KOPANO_TEST_CAPACITY')
    if capacity:
        return int(capacity)

    disptype = user.get(PR_DISPLAY_TYPE_EX)
    capacity = user.get(PR_EMS_AB_ROOM_CAPACITY)

    if disptype == DT_EQUIPMENT and capacity:
        return capacity
    else:
        return 1


def conflict_occurrences(user, item):
    """Item occurrences which overlap (too much) with calendar.

    Args:
        user (User): kopano user object.
        item (Item): kopano item object.

    Returns:
        List[Item]: conflicts list.
    """
    start = item.start + timedelta(days=-1)
    end = start + timedelta(RECURRENCE_AVAILABILITY_RANGE)

    # Determine occurrences which might conflict.
    item_occs = list(item.occurrences(start, end))
    cal_item = item.meetingrequest.calendar_item
    cal_occs = [
        occ for occ in user.calendar.occurrences(start, end) if occ.item != cal_item
    ]
    user_capacity = capacity(user)
    occs = cal_occs + item_occs

    conflicts = []

    # Check based on the datetime range.
    for occ in occs:
        if occ.item is item:
            continue

        # Check time range.
        if (item.start > occ.end and item.end > occ.end) or \
           (item.start < occ.start and item.end < occ.start):
            continue
        conflicts.append(occ)

        # Check capacity threshold.
        if len(conflicts) >= user_capacity:
            break

    if len(conflicts) >= user_capacity:
        return [occ.item for occ in conflicts]
    return []


def conflict_message(occurrences):
    lines = ['The requested time slots are unavailable on the following dates:', '']
    for occ in occurrences:
        lines.append('%s - %s' % (occ.start, occ.end))
    return '\n'.join(lines)


def accept(username, server, entryid=None):
    user = server.user(username)
    autoaccept = user.autoaccept

    if entryid:
        items = [user.item(entryid)]
    else:
        items = []
        for item in user.inbox:
            mr = item.meetingrequest
            if (mr.is_request and mr.response_requested and not mr.processed) or (mr.is_cancellation and not mr.processed):
                items.append(item)

    for item in items:
        mr = item.meetingrequest

        if mr.is_request:
            decline_message = None

            if not autoaccept.recurring and item.recurring:
                decline_message = "Recurring meetings are not allowed"

            elif not autoaccept.conflicts:
                conflicts = conflict_occurrences(user, item)
                if conflicts:
                    decline_message = conflict_message(conflicts)

            if decline_message:
                mr.decline(message=decline_message)
            else:
                mr.accept(add_bcc=True)

        elif mr.is_cancellation:
            mr.process_cancellation(delete=True)

    now = datetime.now()
    try:
        user.freebusy.publish(now - timedelta(7), now + timedelta(180))
    except kopano.errors.NotFoundError as e:
        print("Unable to publish freebusy information: " + str(e), file=sys.stderr)


def main():
    args = sys.argv[1:]
    username, config_file = args[:2]

    entryid = args[2] if len(args) > 2 else None
    config = kopano.Config(filename=config_file)
    server = kopano.server(config=config)

    accept(username, server, entryid)


if __name__ == '__main__':  # pragma: no cover
    main()



# SPDX-License-Identifier: AGPL-3.0-only

from MAPI import (WrapCompressedRTFStream, PT_ERROR, MAPI_MODIFY, MAPI_E_NOT_FOUND,
                  MAPI_E_NOT_ENOUGH_MEMORY, STREAM_SEEK_END)
from MAPI.Util import PROP_TYPE
from MAPI.Tags import (PR_NULL, PR_BODY_W, PR_RTF_COMPRESSED, PR_HTML,
                       PR_RTF_IN_SYNC, PR_USER_ENTRYID, PR_EC_COMPANY_NAME_W,
                       PR_INTERNET_CPID,
                       IID_IStream)

from plugintemplates import MP_CONTINUE, IMapiSpoolerPlugin

import os
import codecs


class Disclaimer(IMapiSpoolerPlugin):

    disclaimerdir = '/etc/kopano/disclaimers'

    def bestBody(self, message):
        tag = PR_NULL
        bodytag = PR_BODY_W  # todo use flags to support PR_BODY_A
        props = message.GetProps([bodytag, PR_HTML, PR_RTF_COMPRESSED, PR_RTF_IN_SYNC], 0)

        if (props[3].ulPropTag != PR_RTF_IN_SYNC):
            return tag

        if((props[0].ulPropTag == bodytag or (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and
           (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and props[1].Value == MAPI_E_NOT_FOUND) and
           (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and props[2].Value == MAPI_E_NOT_FOUND)):
            tag = bodytag
        elif((props[1].ulPropTag == PR_HTML or (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and props[1].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and
             (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY) and
             (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and props[2].Value == MAPI_E_NOT_ENOUGH_MEMORY) and
             not props[3].Value):
            tag = PR_HTML
        elif((props[2].ulPropTag == PR_RTF_COMPRESSED or (PROP_TYPE(props[2].ulPropTag) == PT_ERROR and props[2].Value == MAPI_E_NOT_ENOUGH_MEMORY)) and
             (PROP_TYPE(props[0].ulPropTag) == PT_ERROR and props[0].Value == MAPI_E_NOT_ENOUGH_MEMORY) and
             (PROP_TYPE(props[1].ulPropTag) == PT_ERROR and props[1].Value == MAPI_E_NOT_FOUND) and
             props[3].Value):
            tag = PR_RTF_COMPRESSED

        return tag

    def getCharSetByCP(self, codepage):
        cp2char = {20106:        "DIN_66003",
                   20108:        "NS_4551-1",
                   20107:     "SEN_850200_B",
                   950:          "big5",
                   50221:      "csISO2022JP",
                   51932:           "euc-jp",
                   51936:           "euc-cn",
                   51949:           "euc-kr",
                   949:           "euc-kr",
                   949:            "cp949",
                   949:   "ks_c_5601-1987",
                   936:          "gb18030",
                   936:           "gb2312",
                   936:           "GBK",
                   52936:         "csgb2312",
                   852:           "ibm852",
                   866:           "ibm866",
                   50220:      "iso-2022-jp",
                   50222:      "iso-2022-jp",
                   50225:      "iso-2022-kr",
                   1252:     "windows-1252",
                   1252:       "iso-8859-1",
                   28591:       "iso-8859-1",
                   28592:       "iso-8859-2",
                   28593:       "iso-8859-3",
                   28594:       "iso-8859-4",
                   28595:       "iso-8859-5",
                   28596:       "iso-8859-6",
                   28597:       "iso-8859-7",
                   28598:       "iso-8859-8",
                   28598:     "iso-8859-8-i",
                   28599:       "iso-8859-9",
                   28603:      "iso-8859-13",
                   28605:      "iso-8859-15",
                   20866:           "koi8-r",
                   21866:           "koi8-u",
                   932:        "shift-jis",
                   1200:          "unicode",
                   1201:       "unicodebig",
                   65000:            "utf-7",
                   65001:            "utf-8",
                   1250:     "windows-1250",
                   1251:     "windows-1251",
                   1253:     "windows-1253",
                   1254:     "windows-1254",
                   1255:     "windows-1255",
                   1256:     "windows-1256",
                   1257:     "windows-1257",
                   1258:     "windows-1258",
                   874:      "windows-874",
                   20127:         "us-ascii"}

        return cp2char.get(codepage, "us-ascii")

    def getDisclaimer(self, extension, company):
        if not company:
            company = 'default'

        name = os.path.join(self.disclaimerdir, company + '.' + extension)

        self.logger.logDebug("*--- Open disclaimer file '%s'" % (name))

        return open(name, 'rb').read()

    def PreSending(self, session, addrbook, store, folder, message):

        if not os.path.isdir(self.disclaimerdir):
            self.logger.logWarn("!--- Disclaimer directory '%s' doesn't exist." % self.disclaimerdir)
            return MP_CONTINUE,

        company = None

        props = store.GetProps([PR_USER_ENTRYID], 0)
        if props[0].ulPropTag == PR_USER_ENTRYID:
            currentuser = session.OpenEntry(props[0].Value, None, 0)

            userprops = currentuser.GetProps([PR_EC_COMPANY_NAME_W], 0)
            if userprops[0].ulPropTag == PR_EC_COMPANY_NAME_W and len(userprops[0].Value) > 0:
                company = userprops[0].Value.encode("utf-8")
                self.logger.logDebug("*--- Company name is '%s'" % (company))

        bodytag = self.bestBody(message)

        self.logger.logDebug("*--- The message bestbody 0x%08X" % bodytag)
        if bodytag == PR_BODY_W:

            disclaimer = u"\r\n" + codecs.decode(self.getDisclaimer('txt', company), 'utf-8')

            bodystream = message.OpenProperty(PR_BODY_W, IID_IStream, 0, MAPI_MODIFY)
            bodystream.Seek(0, STREAM_SEEK_END)
            bodystream.Write(disclaimer.encode('utf-32-le'))
            bodystream.Commit(0)

        elif bodytag == PR_HTML:
            charset = "us-ascii"
            props = message.GetProps([PR_INTERNET_CPID], 0)
            if props[0].ulPropTag == PR_INTERNET_CPID:
                charset = self.getCharSetByCP(props[0].Value)

            disclaimer = u"<br>" + codecs.decode(self.getDisclaimer('html', company), 'utf-8')

            stream = message.OpenProperty(PR_HTML, IID_IStream, 0, MAPI_MODIFY)
            stream.Seek(0, STREAM_SEEK_END)
            stream.Write(disclaimer.encode(charset))
            stream.Commit(0)

        elif bodytag == PR_RTF_COMPRESSED:
            self.logger.logWarn("!--- RTF disclaimer is not supported")
            # RTF not supported because this cause body issues.
            return MP_CONTINUE,

            rtf = self.getDisclaimer('rtf', company)

            stream = message.OpenProperty(PR_RTF_COMPRESSED, IID_IStream, 0, MAPI_MODIFY)
            uncompressed = WrapCompressedRTFStream(stream, MAPI_MODIFY)

            # Find end tag
            uncompressed.Seek(-5, STREAM_SEEK_END)
            data = uncompressed.Read(5)
            for i in range(4, 0, -1):
                if data[i] == '}':
                    uncompressed.Seek(i-5, STREAM_SEEK_END)
                    break

            uncompressed.Write(rtf)
            uncompressed.Commit(0)
            stream.Commit(0)
        else:
            self.logger.logWarn("!--- No Body exists")

        return MP_CONTINUE,



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import time
import datetime
import sys

from MAPI.Tags import (
    PR_EC_OUTOFOFFICE, PR_EC_OUTOFOFFICE_SUBJECT_W,
    PR_EC_OUTOFOFFICE_MSG_W, PR_EC_OUTOFOFFICE_FROM,
    PR_EC_OUTOFOFFICE_UNTIL
)
from MAPI.Struct import SPropValue
from MAPI.Time import unixtime

from .errors import NotFoundError

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

class OutOfOffice(object):
    """OutOfOffice class

    Manage out-of-office settings, such as status, subject and message.
    """
    def __init__(self, store):
        self.store = store

    @property
    def enabled(self):
        """Out-of-office is enabled."""
        try:
            return self.store.prop(PR_EC_OUTOFOFFICE).value
        except NotFoundError:
            return False

    @enabled.setter
    def enabled(self, value):
        self.store.mapiobj.SetProps([SPropValue(PR_EC_OUTOFOFFICE, value)])
        _utils._save(self.store.mapiobj)

    @property
    def subject(self):
        """Out-of-office subject."""
        try:
            return self.store.prop(PR_EC_OUTOFOFFICE_SUBJECT_W).value
        except NotFoundError:
            return ''

    @subject.setter
    def subject(self, value):
        if value is None:
            self.store.mapiobj.DeleteProps([PR_EC_OUTOFOFFICE_SUBJECT_W])
        else:
            self.store.mapiobj.SetProps(
                [SPropValue(PR_EC_OUTOFOFFICE_SUBJECT_W, str(value))])
        _utils._save(self.store.mapiobj)

    @property
    def message(self):
        """Out-of-office message."""
        try:
            return self.store.prop(PR_EC_OUTOFOFFICE_MSG_W).value
        except NotFoundError:
            return u''

    @message.setter
    def message(self, value):
        if value is None:
            self.store.mapiobj.DeleteProps([PR_EC_OUTOFOFFICE_MSG_W])
        else:
            self.store.mapiobj.SetProps(
                [SPropValue(PR_EC_OUTOFOFFICE_MSG_W, str(value))])
        _utils._save(self.store.mapiobj)

    @property
    def start(self):
        """Out-of-office is active starting from the given date."""
        try:
            return self.store.prop(PR_EC_OUTOFOFFICE_FROM).value
        except NotFoundError:
            pass

    @start.setter
    def start(self, value):
        if value is None:
            self.store.mapiobj.DeleteProps([PR_EC_OUTOFOFFICE_FROM])
        else:
            value = unixtime(time.mktime(value.timetuple()))
            self.store.mapiobj.SetProps(
                [SPropValue(PR_EC_OUTOFOFFICE_FROM, value)])
        _utils._save(self.store.mapiobj)

    @property
    def end(self):
        """Out-of-office is active until the given date."""
        try:
            return self.store.prop(PR_EC_OUTOFOFFICE_UNTIL).value
        except NotFoundError:
            pass

    @end.setter
    def end(self, value):
        if value is None:
            self.store.mapiobj.DeleteProps([PR_EC_OUTOFOFFICE_UNTIL])
        else:
            value = unixtime(time.mktime(value.timetuple()))
            self.store.mapiobj.SetProps(
                [SPropValue(PR_EC_OUTOFOFFICE_UNTIL, value)])
        _utils._save(self.store.mapiobj)

    @property
    def period_desc(self): # TODO class Period?
        """English description of out-of-office date range."""
        terms = []
        if self.start:
            terms.append('from %s' % self.start)
        if self.end:
            terms.append('until %s' % self.end)
        return ' '.join(terms)

    @property
    def active(self):
        """Out-of-office is currently active (start <= now < end)."""
        if not self.enabled:
            return False
        now = datetime.datetime.now()
        if self.start and now < self.start:
            return False
        if self.end and now >= self.end:
            return False
        return True

    def __unicode__(self):
        return 'OutOfOffice(%s)' % self.subject

    def __repr__(self):
        return self.__unicode__()

    def update(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)



# SPDX-License-Identifier: AGPL-3.0-only
import sys

from MAPICore import *

from .version import __version__

unicode = False

mod_struct = sys.modules.get('MAPI.Struct')
if mod_struct:
    mod_struct.MAPIError._initialize_errors()
